// cALOHA.cpp: implementation of the cALOHA class.
//
//////////////////////////////////////////////////////////////////////

#include "Aloha.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

extern	cVTime globalClock;///Our global clock!
extern	Timers gTimerManager;


cALOHA::cALOHA(short Id,FILE* fptr,macParameter macParam):cMAC(Id,fptr, macParam)
{
	Idle= true;
	pktSent=false;
	firstSent=false;
	pktToSend.length_=PKTLEN;

	//tweak the following based on how energy efficient u want the mac to be. i.e. dont call it as by default it remains on
	amPower.changeDevState( e_am_on, globalClock.globalTime);
	//the following is done if we want to assume that TDMA doesnt use WU
	wuPower.changeDevState(e_wakeup_off,globalClock.globalTime);
}

cALOHA::~cALOHA()
{
	
}


//the send signal in the state machine. 
bool cALOHA::macSend(Packet pkt){

if(Idle){
	//since it unsync ALOHA 
	//first check if this is the first send
	if(firstSent==true){//if not then behave like slotted ALOHA
		pktToSend=pkt;
		OUTCON(m_myId,"cALOHA::macSend","Data to be transmitted at next slot beggining",3);		
		Idle=false;
		pktSent=false;
		return true;
		}
	else{//this is the first pkt.. sent up slots but also transmit this immediately. 
		firstSent=true;
		Idle=false;
		underwaterPhy.PhySendData(pkt);
		pktSent=true;
		TimerCallback* tcbPtr ;///pointer to insert timers for each node
		EventInfo evInfo;

			
		evInfo.eType_=e_mac_slotTime;
		evInfo.nodeId_=m_myId;				
				
		evInfo.timeLeft_=  (timeTicks) ( ((pkt.length_/CHANNEL_DATA_RATE)+m_macParam.fracDelay*MAX_PROP_DELAY)*SEC_TO_TIMETICK+1);
		
		tcbPtr = new Event(evInfo);	
		gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);
		}
	}
	else{

		m_pktsDropped++;
		return false;
	}
return true;
}



bool cALOHA::macDataRxStart(EventInfo& eventInformation, channelState initState){

//ALOHA... we should really just sleep and not be able to sense any data
return false;


}


bool cALOHA::macDataRxEnd(EventInfo& eventInformation, channelState nextState){

	 cMAC::macDataRxEnd( eventInformation,  nextState) ;
/*if(m_pktsReceived ==100)
	gTimerManager.flushTimerArray();//remove all array so that you can finish this soon. 
	*/
return false;
}


bool cALOHA::macDataTxEnd(EventInfo & eventInformation, channelState initState){
TimerCallback* tcbPtr ;///pointer to insert timers for each node
EventInfo evInfo= eventInformation;

	
	m_pktsSent++;


	Idle=true;//so that we can transmit again
	/****
	Traffic Model: Add a call to nodeTxEnd
	****/
	evInfo.eType_=e_tx_end;
	evInfo.nodeId_=m_myId;
	evInfo.timeLeft_= (timeTicks) (0);

	tcbPtr = new Event(evInfo);

	gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//this should result in NodeTransmitEnd being called 
	///Traffic model input ends
				
return false;//for now.. unless we want to creat all event anew !
}

//The Time out event
bool cALOHA::macTimerExpire(EventInfo & eventInformation){
if(eventInformation.eType_==e_mac_slotTime)
	return slotTime(eventInformation);//since the ST event is just a special type of timer!
return false;
}


bool cALOHA::slotTime(EventInfo& eventInformation){
OUTCON(m_myId,"slotTime", "------",3);
if(pktSent==false && Idle==false){
	underwaterPhy.PhySendData(pktToSend);
	pktSent=true;
}

eventInformation.timeLeft_= (timeTicks) ( ((pktToSend.length_/CHANNEL_DATA_RATE)+m_macParam.fracDelay*MAX_RTT)*SEC_TO_TIMETICK+1);
	return true;
}


//no tone for ALOHA

bool cALOHA::macToneDetected(EventInfo & eventInformation){

	OUTCON(m_myId,"cALOHA::macToneDetected","ERR- no tone should ever be sent/detected in ALOHA",2);		

return false;

}
