// cUTLohi.cpp: implementation of the cUTLohi class.
//
//////////////////////////////////////////////////////////////////////

#include "UTLohi.h"

extern	cVTime globalClock;///Our global clock!
extern	Timers gTimerManager;
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

cUTLohi::cUTLohi(short Id,FILE* fptr,macParameter macParam):cMAC(Id,fptr, macParam)
{
	m_contendCount=0;
	dataCollided=dataCollidedByTone=false;
	m_pktsCollided=m_pktsCollidedByTone=m_pktsReceived=m_pktsDropped=m_pktsSent=m_extraSlot=0;

/***	pre alpha-UTLohi code
CONTEND_TO		=	MAX_RTT ;//cUTLohi
//	CONTEND_TO	= MAX_PROP_DELAY ;//aUTLohi

//*****/
	CONTEND_TO	= m_macParam.fracDelay*MAX_RTT;//alpha fractional delay 
	ANTICIPATE_TO  = 3*CONTEND_TO;//in seconds..This is made large so that there is very little chance of incorrectly transiting to idle from anticipate state
	MAX_RECONTEND_TIME	= 3*TONE_DURATION;//this is the duration from which nodes with queued packets to send wait within before sending a tone
}

cUTLohi::~cUTLohi()
{

}

bool cUTLohi::macSend(Packet pkt){
TimerCallback* tcbPtr;///pointer to insert timers for each node
EventInfo evInfo;


	if(isIdle()){//when the mac can handle transmission
		/**
		Here we set up an event which triggers the mac transmission based on 
		its state machine.
		Notice that we have to take care of the fact that if we have the NAV blocked (i.e. some data being transmitted)
		then we shouldnt attempt transmission as of now 
		**/
		
		switch(m_macState){
			case e_mac_idle://donot immediately send tone. randomly backoff to prevent repetitive bi-directional blindness.
				//done
				startBufferTime=globalClock.globalTime;//TAG:latencymeasure

				evInfo.eType_=e_mac_timeOut;
				evInfo.nodeId_=m_myId;
				if(true==m_macParam.hdlBiDeafness)
					evInfo.timeLeft_= (timeTicks) ( (int)(RAND_UNIFORME*DESYNC_WINDOW)*TONE_DURATION*SEC_TO_TIMETICK)				
/*To handle fairness*/	    			+ (timeTicks)(RAND_UNIFORME*(m_extraSlot*MAX_PROP_DELAY)*SEC_TO_TIMETICK); //we make this max prop delay so that it is fairer
				else
/*To handle fairness*/	evInfo.timeLeft_=(timeTicks)(RAND_UNIFORME*(m_extraSlot*MAX_PROP_DELAY)*SEC_TO_TIMETICK);

				/*To handle fairness*/
				m_extraSlot=0;
				//remove the timer for extraslot
				gTimerManager.RemoveTimer(hdlTimeOut);
				
				/*END-To handle fairness*/
				//select a desync slot to poll the channel again
				//not only do we first choose a different slot, but also a random time within the slot to further increase the probability of saving on 
				//synronized bi-directional blindness.
				tcbPtr = new Event(evInfo);

				hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			
				m_contendCount=0;
				
				m_macState=e_mac_contend_backoff;//change ur state, essential
				m_currPacket=pkt;
				break;
			case e_mac_contend_backoff:
				//done
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", " in contend_backoff: THIS SHOULDNT EVEN HAPPEN!",3);
				break;
			case e_mac_contention:
				//done
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "in contend: THIS SHOULDNT EVEN HAPPEN!",3);
				break;
			case e_mac_anticipating:
				//done
				startBufferTime=globalClock.globalTime;//TAG:latencymeasure

				
				m_currPacket=pkt;
				Idle=false;//now we have a data to send
				m_macState=e_mac_anticipating_busy;
				OUTFILE(outFile,m_myId, "STLOHI::macSend()", "e_mac_anticipating",3);
				break;
			case e_mac_anticipating_busy:
				//done
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "anticipating_busy: THIS SHOULDNT EVEN HAPPEN!",3);
				break;

			case e_mac_receiving_busy:
				//done
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "receiving_busy: THIS SHOULDNT EVEN HAPPEN!",3);
				break;	
			case e_mac_receiving:
				//done
				startBufferTime=globalClock.globalTime;//TAG:latencymeasure

				
				m_currPacket=pkt;
				Idle=false;//now we have a data to send
				m_macState=e_mac_receiving_busy;				
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "e_mac_receiving",3);
				break;	
				
			case e_mac_transmitting:
				//done
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "transmitting: THIS SHOULDNT EVEN HAPPEN!",3);
				break;	
				
			default:
				//done
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "Unknown State - state m/c error",3);
				break;
			
			}
		
		
		//change the state of MAC so that it is no longer IDLE, before returning. 
		Idle=false;
		return true;
		
	}
	else{
		OUTCON(m_myId, "Pkt not sent, MAC blocked","&&&&&&",5);
		m_pktsDropped++;
		return false; //inform the upper layer that data could not be sent
	}
}



/*// phyinterfaceUPDATE
bool cUTLohi::macToneRxStart(EventInfo& eventInformation) {
//this event should go directly to the Phy
//This is not at all noticed by the MAC.. present here due to a kludge in the code
underwaterPhy.PhyToneRxStart(eventInformation);

//also if we had MAC in a receiving state its data might get corrupted so we set the appropriate 
//variable.
if(m_macState==e_mac_receiving ||m_macState==e_mac_receiving_busy)
	dataCollidedByTone=true;
return false;

}
*/

//RTE a valid tone detected on phy layer

/*****
phyinterfaceUPDATE
//bool cUTLohi::macToneRxEnd(EventInfo& eventInformation){
***/
bool cUTLohi::macToneDetected(EventInfo& eventInformation){

TimerCallback* tcbPtr ;///pointer to insert timers for each node
EventInfo evInfo= eventInformation;

//first a sanity flag check i.e. is the tone rx actually on at this time, and has it been on long enough!
if(wuPower.currentState !=e_wakeup_rx ){

	if(wuPower.currentState !=e_wakeup_on ){	
	OUTCON(m_myId,"Tone Detected while not in RX mode","NOT IN ON/LISTEN EITHER!!! POWER CNTRL ERR",2);
	}
	else{
	OUTCON(m_myId,"Tone Detected while not in RX mode","BUT IN ON/LISTEN STATE: confirm why",2);
	}
}
else{	//now we check how long it has been in this state
timeTicks timeInDetectState;
	timeInDetectState= (globalClock.globalTime- wuPower.stateStartTime);
	if(timeInDetectState < (timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK)){
		OUTCON(m_myId,"Tone Detected","WU hasnt stayed in rx mode long enough!!!PWR CNTRL ERR",2);		
	}
/*	else
		cout<<timeInDetectState<<endl;*/
}
/******************
no need to do any kludgy stuff now that the phy layer takes 
//care of all possible collision and non-events
//So if this event is happening, you really did have the 
 tone on the channel for long enough to be detected!!!

 However we do have to account for our data transfer
 strategy which requires that we turn on our data
 receiver for enough time that we can lock in the start symbol
i.e.  CHK_DATA_TIME.
Since we KNOW that this is a contention tone, we can be sure 
that the AM device will turn off after the that wait time
 *******************/
 timeTicks currentTime=globalClock.globalTime;
underwaterPhy.amOn(currentTime);
//now go into the future to change it back to the off state
amPower.changeDevState(e_am_off, currentTime+(timeTicks)(CHK_DATA_TIME*SEC_TO_TIMETICK));


switch(m_macState){//handle this event in the state machine (an 8 state SM)
		case e_mac_idle:
			
			OUTCON(m_myId, "RTE -> e_mac_idle", " You received contention",4);
			/*To handle fairness*/
			m_extraSlot=0;
			//remove the timer
			gTimerManager.RemoveTimer(hdlTimeOut);//Question: Why is this needed? which timer has been set?
			/*END-To handle fairness*/
			
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			evInfo.timeLeft_= (timeTicks) (ANTICIPATE_TO*SEC_TO_TIMETICK);

			tcbPtr = new Event(evInfo);

			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			

			//also change your state
			m_macState=e_mac_anticipating;			
			
			m_contendCount=1;//b/c now atleast one node is contending
			
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "RTE -> mac_idle",3);
			break;
		case e_mac_contend_backoff:
			//done
			OUTCON(m_myId,  "RTE->contend_backoff", "Some body else won contention",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "RTE->contend_backoff",3);
			
			m_macState=e_mac_anticipating_busy;//since u have some data to send, but now have to wait for data to transmit

			///first remove the other Timer inserted when gettgin into the contend_bkoff state
			gTimerManager.RemoveTimer(hdlTimeOut);

			//then behave as you have just transited from idle to anticipating state --but with data to send			
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			evInfo.timeLeft_= (timeTicks) (ANTICIPATE_TO*SEC_TO_TIMETICK);

			tcbPtr = new Event(evInfo);
			m_contendCount=1;//b/c now atleast one node is contending
			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			
			
			break;
		case e_mac_contention:
			//done
			OUTCON(m_myId,  "RTE->contention", "Some body else is also contending",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "RTE->contention",3);
			//that is really all I have to do for this state
			//when TO happens then we do backoff
			m_contendCount++;
			m_macState=e_mac_contention;//keep your state
			
			break;
		case e_mac_anticipating:
			//done
			OUTCON(m_myId,  "RTE->anticipating", "Another node contending",3);
			OUTFILE(outFile,m_myId, "STLOHI::macToneRxEnd()", "e_mac_anticipating",3);
			//that is really all I have to do for this state we increment the contenders and
			//this will be used later for better est of wait time

			m_contendCount++;
			m_macState=e_mac_anticipating;//keep your state
			break;
		case e_mac_anticipating_busy:
			//done
			OUTCON(m_myId,  "RTE->anticipating_busy", "Another node contending",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "e_mac_anticipating_busy",3);
			//we increment the contenders and use later in setting the TO correctly
			m_contendCount++;
			m_macState=e_mac_anticipating_busy;//keep your state
			break;

		case e_mac_receiving_busy:
			//done
			OUTCON(m_myId,  "RTE->receiving_busy", "error- unless ortho tone/data channels",2);			
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "e_mac_receiving_busy",3);
			
			break;	
		case e_mac_receiving:
			//done
			OUTCON(m_myId,  "RTE->receiving", "error- unless ortho tone/data channels",2);			
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "e_mac_receiving",3);
			break;	
			
		case e_mac_transmitting:
			//done
			OUTCON(m_myId,  "RTE->transmitting", "data corruption possible if tone/data channel interfere",4);			
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "e_mac_transmitting",3);
			break;	
			
		default:
			//done
			OUTFILE(outFile,m_myId, "UTLOHI::macToneRxEnd()", "Unknown State - state m/c error",3);
			break;
		
		}



return false;//for now.. unless we want to creat all event anew !
//AFFAN: Return from within each case if another event is added.
}


/* // phyinterfaceUPDATE
bool cUTLohi::macToneTxEnd(EventInfo& eventInformation){
timeTicks duration;

//update the power status to have transmitted the tone
wuPower.consumeEnergy(e_wakeup_tx, (timeTicks) (TONE_DURATION*SEC_TO_TIMETICK));
duration =globalClock.globalTime-underwaterPhy.phyToneTxEnds();///this should be directly targeted to the phy layer. should not be a concern of MAC
/////
In order to handle the case where we are the destination of
a data pkt that got corrupted by our transmission (as opposed
to being stomped on by another tone reception!)
I need to set datcollbytone=true
//////
if(m_macState==e_mac_receiving)
dataCollidedByTone=true;
return false;
}


*/

int cUTLohi::count=0;
/**********
RDS	event:
The way this is made to work is that
MAC is first made to look at the channel state provided by the physical layer. 
Now only if the physical state was clear (e_floor state)
that we can actually read the data header. 
However, with a tone being present, although the data header gets corrupted,
we still transit to a data reciving state assuming that the data wasnt for us. 
In case a data was present (in any form) we dont allow the MAC to process the
data reception (as it should'nt be construed as a different packet, in this case
the new data will simply corrupt bytes of an existing transmission). 
***********/
bool cUTLohi::macDataRxStart(EventInfo& eventInformation, channelState initState){
bool isDestination=false;


if (e_dataOverlap == initState||
	e_dataPresent== initState||
	e_dataToneOverlap== initState){ //i.e. we should not be able to detect this data reception at all
	return false;

}
else if(e_floor== initState||
		( (e_tonePresent ==initState ||e_toneOverlap==initState ) &&
		m_macParam.isToneDataOrthognal==true) ){//only in these cases we can correctly receive a data and its header
	
	/**
	Since the data pkt actually has a tone before it,
	we first set up a tone power consumtion (a)
	then (b) we turn on the am receiver.
	**/
	
	timeTicks toneEndTime = (globalClock.globalTime+(timeTicks)(TONE_DURATION*SEC_TO_TIMETICK));
	//a)
	if(wuPower.currentState == e_wakeup_off){//Sanity flags
		OUTCON(m_myId,"macDataRxStart","WU is OFF while receiving for data !!!PWR CNTRL ERR",2);		
	}
	wuPower.changeDevState(e_wakeup_rx, globalClock.globalTime);

	//I removed the line below as it is not in line with a running assumption that WU takes same energy when it is ON
	// and there is tone OR data present on the channel. i.e. it doesnt go to idle power consumption when data is present
	//as it will be in this case.
	//wuPower.changeDevState(e_wakeup_on,toneEndTime);//turn it back to idle state after the TONE_DURATION  TIME
	//b)
	amPower.changeDevState(e_am_rx,toneEndTime);//turn am to receive state after the TONE_DURATION  TIME ( we are not wainting for listen as it is the same: as Wei said)
					


	if(m_myId==eventInformation.packet_.destId_ ||
		-1==eventInformation.packet_.destId_ ){		
		isDestination=true;
		rxPkt.urDestination=true;
		rxPkt.endOfPktReception = globalClock.globalTime +  (timeTicks)((eventInformation.packet_.length_ /CHANNEL_DATA_RATE)*SEC_TO_TIMETICK);
		}
	else{//in this case we should turn off the data rx after reading the header
	
	amPower.changeDevState(e_am_off,
		(toneEndTime+(timeTicks) (READ_HEADER_TIME*SEC_TO_TIMETICK) ));//turn it back to  off state after the reading the header

	}
	
}
//otherwise since we cant read the header when some energy is on the phy
//layer behave as if the data packet is not meant for u (even if it was)

switch(m_macState){//handle this event in the state machine (an 8 state SM)
		case e_mac_idle:
			//done
			OUTCON(m_myId,  "RDS->idle", "incorrect protocol behavior !!",2);	
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_idle",3);
			m_macState=e_mac_idle;//keep ur state
			break;
		case e_mac_contend_backoff:
			//done
			if(isDestination){
				m_macState=e_mac_receiving_busy;
				OUTCON(m_myId,  "RDS->contend_backoff", "data receiving as dest",3);				
			}
			else{
				m_macState=e_mac_anticipating_busy;
				OUTCON(m_myId,  "RDS->contend_backoff", "data receiving but not dest",3);	

			}			
			
			gTimerManager.RemoveTimer(hdlTimeOut);
			//we exit the state based on the mac Data rx end event that is set by the phy layer!
			//in actual practice we should set a timer for the length of the packet to sleep if we are not the dest			
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_contend_backoff",3);
			break;
		case e_mac_contention:
			//done
			if(isDestination){
				m_macState=e_mac_receiving_busy;
				//we consume the energy over here.. as we stay up for as long as the pkt len of this first packet states
				//I also subtract the energy already accounted for reading the packet header to be consistent
//				amPower.consumeEnergy(e_am_rx, (timeTicks)( ((eventInformation.packet_.length_ /CHANNEL_DATA_RATE)-READ_HEADER_TIME)*SEC_TO_TIMETICK));
				
				OUTCON(m_myId,  "RDS->contend", "data receiving as dest",3);				
			}
			else{
				m_macState=e_mac_anticipating_busy;
				OUTCON(m_myId,  "RDS->contend", "data rx but NOT dest, so we sleep",3);	

			}
			
			
			gTimerManager.RemoveTimer(hdlTimeOut);
			//we exit the state based on the mac Data rx end event that is set by the phy layer!
			//in actual practice we should set a timer for the length of the packet to sleep if we are not the dest
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_contention",3);
			break;
		case e_mac_anticipating:
			//done
			if(isDestination){
				m_macState=e_mac_receiving;
				//we consume the energy over here.. as we stay up for as long as the pkt len of this first packet states
				//I also subtract the energy already accounted for reading the packet header to be consistent
//				amPower.consumeEnergy(e_am_rx, (timeTicks)( ((eventInformation.packet_.length_ /CHANNEL_DATA_RATE)-READ_HEADER_TIME)*SEC_TO_TIMETICK));
				OUTCON(m_myId,  "RDS->anticipating", "data receiving as dest",3);				
			}
			else{
				m_macState=e_mac_anticipating;
				OUTCON(m_myId,  "RDS->anticipating", "data rx but NOT dest, so we sleep",3);		
			}
			
			
			gTimerManager.RemoveTimer(hdlTimeOut);
			//we exit the state based on the mac Data rx end event that is set by the phy layer!
			//in actual practice we should set a timer for the length of the packet to sleep if we are not the dest
			m_contendCount=0;//doing this just to make sure we are not confused
			OUTFILE(outFile,m_myId, "STLOHI::macDataRxStart()", "e_mac_anticipating",3);
			break;
		case e_mac_anticipating_busy:
			//done
			if(isDestination){
				m_macState=e_mac_receiving_busy;
				//we consume the energy over here.. as we stay up for as long as the pkt len of this first packet states
				//I also subtract the energy already accounted for reading the packet header to be consistent
//				amPower.consumeEnergy(e_am_rx, (timeTicks)( ((eventInformation.packet_.length_ /CHANNEL_DATA_RATE)-READ_HEADER_TIME)*SEC_TO_TIMETICK));
				OUTCON(m_myId,  "RDS->anticipating_busy", "data receiving as dest",3);				
			}
			else{
				m_macState=e_mac_anticipating_busy;
				OUTCON(m_myId,  "RDS->anticipating_busy", "data rx but NOT dest, so we sleep",3);		
			}			
			
			gTimerManager.RemoveTimer(hdlTimeOut);
			//we exit the state based on the mac Data rx end event that is set by the phy layer!
			//in actual practice we should set a timer for the length of the packet to sleep if we are not the dest
			m_contendCount=0;//doing this just to make sure we are not confused
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_anticipating_busy",3);
			break;

		case e_mac_receiving_busy:
			//done
			OUTCON(m_myId,  "RDS->receiving_busy", "Data Collision!!",2);			
			/* // phyinterfaceUPDATE
			dataCollided=true;

			*/
			//currently nothing more to do, except to wait for RDE 
			m_macState=e_mac_receiving_busy;//keep state
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_receiving_busy",3);
			break;	
		case e_mac_receiving:
			//done
			OUTCON(m_myId,  "RDS->receiving", "Data Collision!!",2);			

/* // phyinterfaceUPDATE
			dataCollided=true;

			*/
			//currently nothing more to do, except to wait for RDE 	
			m_macState=e_mac_receiving;//keep state
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_receiving",3);
			break;	
			
		case e_mac_transmitting:
			//done
			/***
			in a real system, this will not happen, since if you will have ur receiver off while transmitting!
			**/
			OUTCON(m_myId,  "RDS->transmitting", "blocked by phy layers initState !!",2);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "e_mac_transmitting",3);
			break;	
			
		default:
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxStart()", "Unknown State - state m/c error",3);
			break;
		
		}

return false;
}

/*** 
RDE Event
---------
Read Base class implementation for design insight

Essentially this assumes that receiver of data keep in receiving state (sleeping
if they were not the dest) until all data is cleared on the channel). 

***/

bool cUTLohi::macDataRxEnd(EventInfo& eventInformation, channelState nextState){

timeTicks backoffTime=0;
/***
The idea over here is that we are trying to replicate the behavior of real radios when collisions occur.
So when a radio does get a packet, it will hook on and receive it based on the packet length (PLEN)
provided (hopefully ) in the pkt header. 

Once that PLEN bytes have been received, it will turn off its receiving mechanism. 

However, if there is a data overlap, then even after PLEN bytes, there is data on the channel. 
Ideally we would have nodes transit when PLEN bytes have been received. However, since the orignal
state m/c didnt have this provision, we are using a hack that allows us to keep the same state m/c and
transit only when all the data on the channel has finished. 

The hack is that we assume that once PLEN bytes have been received, nodes that have pkts to send
or not go to a state where they constantly poll the channel to be clear of data. Once it is cleared of data 
i.e. after the RDE that doesnt return false from cMAC::macDataRxEnd(), we transit to the idle or contend
state as described in our state machines.

Now the first part of hack is where we look at whether we are the destination of the currently synced pkt
and whether the RDE is for that pkt. 
The reason we only do this is b/c the only reason we want to know if the pkt we synced to has ended,
given the above hack/assumption, is that we want to turn off the AM to off state. If we are not the dest
then we shouldnt have it turned on any way. 

	***/
if(rxPkt.urDestination == true  && rxPkt.endOfPktReception== globalClock.globalTime){

			if(amPower.currentState == e_am_rx) //only turn off while receiving
				underwaterPhy.amOff(globalClock.globalTime);
			else
				OUTCON(m_myId, "RDE->receiving_busy", "AM not in RX mode! POWER CONTROL ERR",2);
			
			rxPkt.urDestination= false;//make it false so that we dont repeatedly turn amOff. 
	}

if(cMAC::macDataRxEnd( eventInformation,  nextState) == false)//this basically tells if the data has finished on the channel or not
		return false;
//otherwise we process the RDE event as a state machine transition.

	TimerCallback* tcbPtr ;///pointer to insert timers for each node
	EventInfo evInfo;


switch(m_macState){//handle this event in the state machine (an 8 state SM)
		case e_mac_idle:
			//this might happen when the RDE is actually an end of data  resulting b/c of a tx data and receive data overlap
			if(e_dataOverlapEnded !=nextState ){
					OUTFILE(outFile,m_myId, "RDE->idle", "incorrect protocol behavior !!",2);
				}
				else{//do nothing and stay in the same state (as this is really implying that the data channel is now clear- this should be rx at MAC at all)
					m_macState= e_mac_idle;
				}
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_idle",3);
			break;
		case e_mac_contend_backoff:
			//done
//			
			//this might happen when the RDE is actually an end of data .. so 
			if(e_dataOverlapEnded !=nextState ){
				OUTCON(m_myId,  "RDE->contend_backoff", "incorrect protocol behavior !!",2);
			}
			else{//do nothing and stay in the same state
				m_macState= e_mac_contend_backoff;
			}
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_contend_backoff",3);
			break;
		case e_mac_contention:
			//done
			OUTCON(m_myId,  "RDE->contend", "incorrect protocol behavior !!",2);				
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_contention",3);
			break;
		case e_mac_anticipating:
			//done
			OUTCON(m_myId,  "RDE->anticipating", " External Data transfer finished",3);
			//Now go to idle state 
			//also turn on the wu hardware so that we can receive any other contention tones
			underwaterPhy.wakeUpOn(globalClock.globalTime);
			m_macState= e_mac_idle;
			m_contendCount=0;
			
			OUTFILE(outFile,m_myId, "STLOHI::macDataRxEnd()", "e_mac_anticipating",3);
			break;
		case e_mac_anticipating_busy:
			//done
			//now set a random TO and go to contend_bk off state
			//since u already have an packet queued to send but want to wait some time

			//also turn on the wu hardware so that we can receive any other contention tones
			underwaterPhy.wakeUpOn(globalClock.globalTime);

			backoffTime = (timeTicks) (RAND_UNIFORME*MAX_RECONTEND_TIME*SEC_TO_TIMETICK)+1;//select within MAX_RECONTEND_TIME
			//the plus one is to prevent tone following the data at the same instance

			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			evInfo.timeLeft_= backoffTime;

			tcbPtr = new Event(evInfo);

			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			
			m_contendCount=0;
			
			m_macState=e_mac_contend_backoff;//change ur state, essential
			OUTCON(m_myId,  "RDE->anticipating_busy", " External Data transfer finished- now we can contend",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_anticipating_busy",3);
			break;

		case e_mac_receiving_busy:
			//done			
			
			//also turn on the wu hardware
			underwaterPhy.wakeUpOn(globalClock.globalTime);

			//now set a random TO and go to contend_bk off state
			//since u already have an packet queued to send but want to wait some time
			backoffTime = (timeTicks) (RAND_UNIFORME*MAX_RECONTEND_TIME*SEC_TO_TIMETICK)+1;//select within MAX_RECONTEND_TIME
			//the plus one is to prevent tone following the data at the same instance
			
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			evInfo.timeLeft_= backoffTime;

			tcbPtr = new Event(evInfo);

			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			
			m_contendCount=0;
			m_macState = e_mac_contend_backoff;//essential
			//here we decide as a receiver if the pkt we received was corrupted by data or tone or was all clear

			
			OUTCON(m_myId,  "RDE->receiving_busy", " Data transfer finished- now we contend",3);

			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_receiving_busy",3);
			break;	
		case e_mac_receiving:
			//done
			
			//also turn on the wu hardware
			underwaterPhy.wakeUpOn(globalClock.globalTime);

			m_contendCount=0;
			m_macState= e_mac_idle;
			
				
			OUTCON(m_myId,  "RDE->receiving", " Data transfer finished- now we contend",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_receiving",3);
			break;	
			
		case e_mac_transmitting:
			//done
			OUTCON(m_myId,  "RDE->transmitting", " Should not happen",2);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "e_mac_transmitting",3);
			break;	
			
		default:
			OUTFILE(outFile,m_myId, "UTLOHI::macDataRxEnd()", "Unknown State - state m/c error",3);
			break;
		
		}

return false;
}



/*** 
TDE Event
---------
When transmission ends, right now
we dont do anything at all. We consider
everything to be perfect, and simply transit to 
the idle state.

However, we should check the channel status once we attempt again (from the idle
state) and make sure that we dont send while the channel is bad

***/
bool cUTLohi::macDataTxEnd(EventInfo& eventInformation, channelState initState){
//only makes sense if u were in the transmitting state .. otherwise some thing crappy is happening
TimerCallback* tcbPtr ;///pointer to insert timers for each node
EventInfo evInfo;




totalLatency+=globalClock.globalTime-startBufferTime;
switch(m_macState){//handle this event in the state machine (an 8 state SM)
		case e_mac_idle:
			OUTCON(m_myId,  "TDE->e_mac_idle", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_idle",3);
			break;
		case e_mac_contend_backoff:
			OUTCON(m_myId,  "TDE->e_mac_contend_backoff", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_contend_backoff",3);
			break;
		case e_mac_contention:
			OUTCON(m_myId,  "TDE->e_mac_contention", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_contention",3);
			break;
		case e_mac_anticipating:
			OUTCON(m_myId,  "TDE->e_mac_anticipating", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "STLOHI::macDataTxEnd()", "e_mac_anticipating",3);
			break;
		case e_mac_anticipating_busy:
			OUTCON(m_myId,  "TDE->e_mac_anticipating_busy", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_anticipating_busy",3);
			break;

		case e_mac_receiving_busy:
			OUTCON(m_myId,  "TDE->e_mac_receiving_busy", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_receiving_busy",3);
			break;	
		case e_mac_receiving:
			OUTCON(m_myId,  "TDE->e_mac_receiving", " Protocol incorrectness!",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_receiving",3);
			break;	
			
		case e_mac_transmitting:
			//done 
			
			m_macState=e_mac_idle;
			m_contendCount=0;
			///to handle fairness
			m_extraSlot=1;
			//also add a time out after which the extra slot effect should vanish.
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			evInfo.timeLeft_= (timeTicks) MAX_PROP_DELAY*SEC_TO_TIMETICK;

			tcbPtr = new Event(evInfo);

			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			
			m_contendCount=0;
			/////handle Fairness
			m_pktsSent++;

			/****
			Traffic Model: Add a call to nodeTxEnd
			****/
			evInfo.eType_=e_tx_end;
			evInfo.nodeId_=m_myId;
			evInfo.timeLeft_= (timeTicks) (0);

			tcbPtr = new Event(evInfo);

			gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//this should result in NodeTransmitEnd being called 

			//once we have transmitted, surely turn OFF the am modem b/c we dont want to listen to any data what so ever
			underwaterPhy.amOff(globalClock.globalTime);
			//also turn on the wakeup receiver
			underwaterPhy.wakeUpOn(globalClock.globalTime);
			
			OUTCON(m_myId,  "TDE->transmitting", " Data transfer finished",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "e_mac_transmitting",3);
			break;	
			
		default:
			OUTFILE(outFile,m_myId, "UTLOHI::macDataTxEnd()", "Unknown State - state m/c error",3);
			break;
		
		}

return false;
}

bool cUTLohi::macTimerExpire(EventInfo& eventInformation){
TimerCallback* tcbPtr;///pointer to insert timers for each node
EventInfo evInfo;
 

switch(m_macState){//handle this event in the state machine (an 8 state SM)
		case e_mac_idle:
			
			///fairness issue
			m_extraSlot=0;
			///fairness done
			OUTCON(m_myId,  "TO->idle", " Need to reset the fairness parameter",3);
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "e_mac_idle",3);
			break;
		case e_mac_contend_backoff:{
				//done
				channelState phyState= underwaterPhy.getDataChannelState();
				/******* 
				phyInterfaceUPDATE 
				******/
				if(phyState== e_floor){//i.e. when the channel has NO form of energy
				OUTCON(m_myId,  "TO->contend_backoff", "Channel not busy, sending tone",3);
				OUTFILE(outFile,m_myId, "UTLOHI::macSend()", "e_mac_idle",3);
				underwaterPhy.PhySendTone();

				evInfo.eType_=e_mac_timeOut;
				evInfo.nodeId_=m_myId;				
			
				evInfo.timeLeft_= (timeTicks) (CONTEND_TO*SEC_TO_TIMETICK);
	
				tcbPtr = new Event(evInfo);
	
				hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			

				
				m_macState=e_mac_contention;
				m_contendCount=1;//b/c now atleast you are contending				

				Idle=false;//we have gone into one of the starred states where any mac send will be blocked
				}
				else{//when channel has ANY form of energy
				evInfo.eType_=e_mac_timeOut;
				evInfo.nodeId_=m_myId;
				evInfo.timeLeft_= 
				evInfo.timeLeft_= (timeTicks) ( RAND_UNIFORME*DESYNC_WINDOW*TONE_DURATION*SEC_TO_TIMETICK)+(TONE_DURATION*SEC_TO_TIMETICK);
				
				//actually slot any time within a random duration to poll the channel again to see if it is clean
				//but at the minimum wait for one tone_duration time! (nothing to do with tone time)
				
				//TODO:: Look at this, is there a better way of doing it?
				tcbPtr = new Event(evInfo);

				hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed			
				m_contendCount=0;
				
				m_macState=e_mac_contend_backoff;//change ur state, essential
				}
			break;
		}
		case e_mac_contention:
			//done
			if(1==m_contendCount){//this means u have won the contention .. transmit!!!
				
				/**
				update the power  of wu.
				NOTE:i account for the tone that  i (theoratically) should have transmitted before the data pkt
				**/
				OUTCON(m_myId,  "TO->contend", " Contention won.. Tx'ing",3);
				
//				wuPower.consumeEnergy(e_wakeup_tx, (timeTicks)(TONE_DURATION*SEC_TO_TIMETICK) );//doing this inside the phy layer now
				
				//first we increase the length of the pkt by the TONE DURATION (in terms of # bytes)
				m_currPacket.length_+=(uint_32) (TONE_DURATION*CHANNEL_DATA_RATE);				
				underwaterPhy.PhySendData(m_currPacket);//send out the current data pkt - 
				/*
				the power of pkt tx is updated inside the phy layer
				However since the pkt is longer by the tone tx we have to:
				a) reduce the am.tx time by that amount
				b) add wu.tx cost
				*/
				//a)
				timeTicks toneTxEndTime =(globalClock.globalTime+(timeTicks)(TONE_DURATION*SEC_TO_TIMETICK));

				amPower.changeDevState(e_am_off,globalClock.globalTime);//turn it off as starters
				amPower.changeDevState(e_am_tx,toneTxEndTime);//then turn it on at the correct time in the future
				//b)
				wuPower.changeDevState(e_wakeup_tx,globalClock.globalTime);//turn it to tx state 
				wuPower.changeDevState(e_wakeup_on, toneTxEndTime);//turn it to idle state at the correct time
				
				m_macState=e_mac_transmitting;
				m_contendCount=0;			

			}
			else{//otherwise there were other contenders... choose contending slot intelligently
				OUTCON(m_myId,  "TO->contend", " Contention collided.. backingoff",3);
			
				int contend_slot= (int)(RAND_UNIFORME*m_contendCount); //randomly choose some amount of time to wait
				/***
				Observe that if you wake up 3 contention slots later and multiple contentions have
				already happened, you will over count the # of contenders.. 
				However, this naturally results in lesser contention and helps 
				in getting a data contention asap. 
				***/
				//now setup a timer for knowing if u won the contention or not
				evInfo.eType_=e_mac_timeOut;
				evInfo.nodeId_=m_myId;

				evInfo.timeLeft_= (timeTicks) (contend_slot*CONTEND_TO*SEC_TO_TIMETICK)+
						(timeTicks) ( (double)(RAND_UNIFORME*DESYNC_WINDOW)*TONE_DURATION*SEC_TO_TIMETICK);
		
				tcbPtr = new Event(evInfo);
		
				hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed	
				m_macState=e_mac_contend_backoff;
				m_contendCount=0;

			}
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "e_mac_contention",3);
			break;
		case e_mac_anticipating:
			//done
			if(0==m_contendCount){//this means no one contended in the last anticipation period (very small probability)
			//in this case, it better to transit to idle rather than be stuck for ever in anticipation
			m_macState=e_mac_idle;
			OUTCON(m_myId,  "TO->anticipating", " Timed out waiting in Anticipate state",3);
			}
			else{//in the other case, we wait for a time proportional to the # of contenders..
			m_macState=e_mac_anticipating;//and stay in this state
			//now setup a timer 
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			m_contendCount++;//this is done to cater for any deafness that might have occured... additive increase 
			evInfo.timeLeft_= (timeTicks) (CONTEND_TO*SEC_TO_TIMETICK*m_contendCount);
	
			tcbPtr = new Event(evInfo);
	
			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed	
			m_contendCount=0;//also reset the contend count!
			OUTCON(m_myId,  "TO->anticipating", " Setting another Timer",3);
			}
			
			OUTFILE(outFile,m_myId, "STLOHI::macTimerExpire()", "e_mac_anticipating",3);
			break;
		case e_mac_anticipating_busy:
			//done
			gTimerManager.RemoveTimer(hdlTimeOut);
			if(0==m_contendCount){//this means no one contended in the last anticipation period (very small probability)
			//in this case, it better to attempt contention (as we already have a data to send) than to be stuck forever
			
				
			//now setup a timer after which u will attempt contention - in this time you are "sensing" 
			//for another person's contention
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;

			//now u already have an packet queued to send but want to wait some time
			//wait this much before attempting contention
			evInfo.timeLeft_= (timeTicks) (RAND_UNIFORME*MAX_RECONTEND_TIME*SEC_TO_TIMETICK);//select within MAX_RECONTEND_TIME
	
			tcbPtr = new Event(evInfo);
	
			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed	
			m_macState=e_mac_contend_backoff;

			OUTCON(m_myId,  "TO->anticipating_busy", " Timed out now attempting contention",3);
			}
			else{//in the other case, we wait for a data transfer for a time proportional to the # of contenders..
			m_macState=e_mac_anticipating_busy;//and stay in this state
			//now setup a timer 
			evInfo.eType_=e_mac_timeOut;
			evInfo.nodeId_=m_myId;
			m_contendCount++;//this is done to cater for any deafness that might have occured... additive increase 
			evInfo.timeLeft_= (timeTicks) (CONTEND_TO*SEC_TO_TIMETICK*m_contendCount);
	
			tcbPtr = new Event(evInfo);
	
			hdlTimeOut=gTimerManager.AddTimer(evInfo.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed	
			m_contendCount=0;//also reset the contend count!
			
			OUTCON(m_myId,  "TO->anticipating_busy", " Setting another Timer",3);
			}
			
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "e_mac_anticipating_busy",3);
			break;

		case e_mac_receiving_busy:
			OUTCON(m_myId,  "TO->receiving_busy", " Protocol incorrectness!",2);
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "e_mac_receiving_busy",3);
			break;	
		case e_mac_receiving:
			OUTCON(m_myId,  "TO->receiving", " Protocol incorrectness!",2);			
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "e_mac_receiving",3);
			break;	
			
		case e_mac_transmitting:
			OUTCON(m_myId,  "TO->e_mac_transmitting", " Protocol incorrectness!",2);
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "e_mac_transmitting",3);
			break;	
			
		default:
			OUTFILE(outFile,m_myId, "UTLOHI::macTimerExpire()", "Unknown State - state m/c error",3);
			break;
		
		}

return false;
}
