/*
 * Author: Affan Syed <asyed@usc.edu>
 *   channel.cpp: The channel propagation and noise model is implemented as a shared class 
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */


#include "channel.h"


extern	cVTime globalClock;///Our global clock!
extern	vector<nodeInfo> gNodeInfoVector;///vector that will keep track of all the nodes in the simulation
extern	Timers gTimerManager;


bool cChannel::sendTone(Packet pktToSend){

	TimerCallback* tcbPtr;///pointer to insert timers for each node
	int nodeIter=0;
	int nodeCount = gNodeInfoVector.size();
	double distance=0;
	
	vector<nodeInfo>::iterator nodePosition;
	

	nodePosition = find(gNodeInfoVector.begin(), gNodeInfoVector.end(),pktToSend.sourceId_);
	
	if(nodePosition == gNodeInfoVector.end()) { ///we didnt find the node id ..something fishy!
	
		cerr<<"node id "<<pktToSend.sourceId_<<" Could not be found .cChannel::sendTone("<<endl;
		
		
		return false;
	}
	
	/*****
	Set the values for the event 				
	***/			
	timerInfo.eType_ = e_tone_rx_start;//set up an application layer transmission event

	//now we loop over all the nodes in the node vector and set up a reception time based on the distance b/w nodes
	
	for (nodeIter=0;nodeIter<nodeCount; nodeIter++){
		if(gNodeInfoVector[nodeIter].myId != pktToSend.sourceId_){
			timerInfo.nodeId_=gNodeInfoVector[nodeIter].myId;//set the event for this node

			distance = (*nodePosition).getDistance(gNodeInfoVector[nodeIter]);
			
			timerInfo.timeLeft_ = (timeTicks)( distance*SEC_TO_TIMETICK/(SNDSPEED_MPS));
						
			timerInfo.packet_ = pktToSend;
			//AFFAN: No need to add the rssi value for tone.. assuming that it works is a binary channel anyway

		//	cout<<"Sending Tone from "<<pktToSend.sourceId_<< " To "<<timerInfo.nodeId_<<" at distance "<<distance<<endl;
			tcbPtr = new Event(timerInfo);
			//insert the timer into the timer manager
			gTimerManager.AddTimer(timerInfo.timeLeft_, tcbPtr);
			}
		}
return true;

}
bool cChannel::sendData(Packet pktToSend){

	TimerCallback* tcbPtr;///pointer to insert timers for each node
	int nodeIter=0;
	int nodeCount = gNodeInfoVector.size();
	double distance=0;
	
	vector<nodeInfo>::iterator nodePosition;
	

	nodePosition = find(gNodeInfoVector.begin(), gNodeInfoVector.end(),pktToSend.sourceId_);
	
	if(nodePosition == gNodeInfoVector.end()) { ///we didnt find the node id ..something fishy!
	
		cerr<<"node id "<<pktToSend.sourceId_<<" Could not be found .cChannel::sendTone("<<endl;
		
		
		return false;
	}
	
	/*****
	Set the values for the event 				
	***/			
	timerInfo.eType_ = e_data_rx_start;//set up an application layer transmission event

	//now we loop over all the nodes in the node vector and set up a reception time based on the distance b/w nodes
	
	for (nodeIter=0;nodeIter<nodeCount; nodeIter++){
		if(gNodeInfoVector[nodeIter].myId != pktToSend.sourceId_){
			timerInfo.nodeId_=gNodeInfoVector[nodeIter].myId;//set the event for this node

			distance = (*nodePosition).getDistance(gNodeInfoVector[nodeIter]);
			if(distance > MAX_RANGE)//basically dont send to nodes that are more than MAX_RANGE AWAY
				continue;
			timerInfo.timeLeft_ = (timeTicks) (distance*SEC_TO_TIMETICK/(SNDSPEED_MPS));
			
			
			timerInfo.packet_ = pktToSend;
		//	cout<<"Sending data from "<<pktToSend.sourceId_<< " To "<<timerInfo.nodeId_<<" at distance "<<distance<<endl;
			//TODO:: Add the rssi value based on some equation like so
			//timerInfo.packet_.rssi= DEFAULT_TRANSMIT_POWER/distance^CHANNEL_LOSS_EXPONENT
			
			tcbPtr = new Event(timerInfo);
			//insert the timer into the timer manager
			gTimerManager.AddTimer(timerInfo.timeLeft_, tcbPtr);
			}
		}
return true;

}
cChannel gChannel;
