/*
 * Author: Affan Syed <asyed@usc.edu>
 *   defines.h: All general purpose defines/enums strucs and declarations made over here
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */

#ifndef _DEFINES_H_
#define _DEFINES_H_


#include "timers.h"
#include "randomNumber.h"


#define MAX_TIMER_VALUE    0xfffffff // no timer present 

//defines for the simulators

 #define MAX_X 400
#define MAX_Y 300 //this keeps the max internode distance at 500m !


/* Miscellaneous constants, defines and macros*/
#define   CLOCK_GRANULARITY 1 //granularity of the clock in terms of timeTicks. So the time t node clocks can report is (int)t/CG *CG

#define   SNDSPEED_MPS			1500// base sound speed in m/s
#define   SNDSPEED_RANGE		60//max variation in sound from 1470-1530
#define   TIMETICK_TO_USEC		100///how much is microsec equal 1 time tick (atomic unit of our simulator)
#define   SEC_TO_TIMETICK      	((double)1000000/TIMETICK_TO_USEC)//how many time ticks equal 1 second


#define   PKTLEN					81//constant or MAX pkt length (bytes) of data that is transmitted by application layer 
#define   PKT_HDR			   	5 //5 bytes used for pkt hdr len
#define   PAMBLE_START_SYMBL 	5// not sure have to confirm with wei or jack (# of bytes for preamble + start symbol)


#define TONE_DURATION			0.005 //TONE duration in sec
#define TONE_DETECT_TIME		TONE_DURATION//the time required to detect the tone
#define CHANNEL_DATA_RATE 		((double)125) //the data transmit rate for the channel (in Bytes/s)

#define READ_HEADER_TIME		((double) (PKT_HDR/CHANNEL_DATA_RATE))
#define CHK_DATA_TIME			((double) (PAMBLE_START_SYMBL/CHANNEL_DATA_RATE))
#define BURST_PARAM			(double) 10 /// the parameter defining the burstiness of traffic if used


/*
#define 	MAX_RANGE			1500 //in meters
#define   MAX_PROP_DELAY		((double)MAX_RANGE/SNDSPEED_MPS )//in seconds
#define  	MAX_RTT				2*MAX_PROP_DELAY//in seconds
*/
#define DESYNC_WINDOW			2// this is the # of slots over which a node will choose to transmit so that it isnt sync

//this macro is to be used only when assigning events for NODE specific issues, i.e. not for calculating the TOF
#define granularityConversion(t) (  ((int) (t/CLOCK_GRANULARITY))*CLOCK_GRANULARITY)
#define DEBUG_LEVEL_CON		0 //higher debug values are of less importance
#define DEBUG_LEVEL_FILE	0 //higher debug values are of less importance
#define OUTFILE(fptr,id,eventStr,detailStr,dbg) if( dbg<= DEBUG_LEVEL_FILE) fprintf(fptr,"%ld\t %d\t %s \t %s\n",globalClock.globalTime,id,eventStr,detailStr);
#define OUTCON(id,eventStr,detailStr,dbg) if (dbg<= DEBUG_LEVEL_CON) printf("%ld\t %d\t %s \t %s\n",globalClock.globalTime,id,eventStr,detailStr);



#define	min(a,b)	((a) < (b) ? (a) : (b))
#define	max(a,b)	((a) > (b) ? (a) : (b))
#define   absolute(a) ((a) < 0 ? -a : a)

#define uint_32 unsigned int 
#define uint_16 unsigned short
#define uint_8   unsigned char


enum channelState{
	e_floor,//with no transmission on the channel 
	e_dataPresent,// when channel at node has a single data transmission
	e_dataOverlap,//when channel at node is or was corrupted by data overlap
	e_tonePresent,//when channel at node has a single tone;
	e_toneOverlap,//when channel at node hasor had multiple tone;
	e_dataToneOverlap,//when channel at node has/had both tone and data before having returned to floor
	e_dataOverlapEnded //this is abstract only when a previous RDE has made the data channel have only 
	//a sinle data source (that was previously corrupted) - we never transit to this state
};
 
enum MsgType {
				 e_msg_tone,
				 e_msg_data
				 };

				 
enum MacType {
				 e_UTLohi, 
				 e_STLohi,
				 e_TDMA,
				 e_CSMA,
				 e_ALOHA,
				 e_SALOHA
				 }; //add any more MAC types over here.

enum PktType {
				 e_constant, //same pkt every time
				 e_uniform //vary pkt lenth uniformly over pkt length
				 }; 
//These are the event types that decide what callback function to be notified.
enum eventType {e_tx_end, //for application layer reception
				   e_tx,//for application layer transmit event (traffic generation)
				   //Tone channel events
				   e_tone_rx_start,
				   e_tone_rx_end,//both set events of when mac receive starts or ends
/*This might not be required*/ //  e_tone_tx_mac_start,
				   e_tone_tx_end,//both set events of when mac transmit starts or ends
				   //data channel events
				   e_data_rx_start,
				   e_data_rx_end,//both set events of when mac receive starts or ends
/*This might not be required*/ //e_data_tx_mac_start,
				   e_data_tx_end,//both set events of when mac transmit starts or ends
				   e_mac_timeOut, //for a general timer that can be set
				   e_mac_slotTime, //signals the synchronized time for transmitting in the slotted case
				   e_mac_toneDetected,//this signal comes from the phy when tone has stayed, uninterrupted
				   					  //on the data channel for TONE_DETECT time
/** these are extra commands for un real communication between nodes*/
				    e_abstract_reportLoss, //signals that pkt sent by the node receiving this event was corrupted by data at the receiver. 
				    e_abstract_reportLossByTone	
};
enum trafficType{e_manual, //with this traffic leave it to the application to generate all traffic types
				e_poisson, //inter arrival time b/w pkts is poisson distributed.
				e_CBR,//traffic is generated after every specific interval. 
				e_bursty,//bursty traffic with average poisson rate defined.. burst scale of 1/10
				e_saturated//when the node will always have a packet to send... 
};
struct trafficInfo{
	trafficType eType;
	double parameter;//generic traffic parameter
	//lamda for the posson
	//delta time for CBR (after which the new traffic will be generated).
	PktType ePktType;
	long pktLength;
};
enum nodeState {
				   e_app_transmiting,
				   e_app_receiving,
				   e_app_idle, //all three describe the applicatin layer state
				   /****
				   Common   MAC STATES
				   ****/
				   e_mac_idle,
  				   e_mac_contend_backoff,//a sort of go between state -refer to state m/c for details
				   e_mac_anticipating,//where a node is passive with no data to send of its own
				   e_mac_receiving,//data w/o any data to send
				   e_mac_contention,
				   e_mac_transmitting,	
				   /***
				   STATES for the UTLohi MAC only
				   ***/			   				   
				   e_mac_anticipating_busy,// node made to be passive (due to other contention), but with data to send				   
				   e_mac_receiving_busy,//data receiving, but now with data to send after this				   
				   /****
				   Some additional state for STLohi MAC
				   ****/
				   e_mac_wait_slot,//here a node waits for a the correct time before attempting contention
				   e_mac_wait_data,//here a node is waiting for an external data transfer to end. 
				   e_mac_wait_desync//this is to ensure that nodes bi-directional deafness is not repeated
				   
	};

/****
Power related defiens

****/

#define WAKEUP_TX_POWER 2000;
#define WAKEUP_OFF_POWER 0.5;
#define WAKEUP_RX_POWER 0.5;//assuming for now that the wake up rx and sleep taek the same power

//AM stands for Acoustic Modem
#define AM_TX_POWER (2000);
#define AM_OFF_POWER 0.5;
#define AM_RX_POWER 20;

enum wakeupStates {
	e_wakeup_tx,
	e_wakeup_on,
	e_wakeup_rx,
	e_wakeup_off
};


enum acousticModemStates {
	e_am_tx,
	e_am_on,
	e_am_rx,
	e_am_off
};




struct Coordinate { ///coordinates in meters!
	float x,y;
	Coordinate(float X,float Y){
		x=X;
		y=Y;
		}
	Coordinate(){
		x=-1;//make sure the coordinates are
		y=-1;//invalid at initialization
		}
};

struct nodeInfo{//all channel related info needed for a node

short myId;
Coordinate myLocation;
nodeInfo(short Id,Coordinate location){
	myId=Id;
	myLocation=location;
	};
//This func is used to compute the euclidean distance b/w two nodes
double getDistance(const nodeInfo &node ){
  
  double first=0,second=0;

  first =  (myLocation.x-node.myLocation.x);
  second = (myLocation.y-node.myLocation.y);

  
  return sqrt(first * first + second* second);
};
bool operator==(const short id) const { return (this->myId ==id);}; //this is used in the STL find function


};


struct macParameter {
	MacType myMacType;
	bool hdlBiDeafness;
	bool isToneDataOrthognal;//defines whether the phy should consider tone and data to be on separate channels or not
	double maxRange;//in meters what is the maximum range of the node
	float fracDelay;//used as 1)  the frac of delay used to increase the slot length of saloha - for the WUWNET submission
	float toneCancelProb;//prob that a tone is dropped/loss due to channel noise
	// 2) define the fraction of MAX_RTT that the UTLohi will have its contention round equal to. (extention to SENSYS/INFOCOM submission)
};
//Container that might be either tone or packet
struct Packet{
	MsgType msgType_;//tone or data (for now)
	short sourceId_;//the transmitting node
	short destId_; //if -1 implies a broadcast!- otherwise tells the node id of the intended receiver. 
	uint_32 length_;//size of the packet in terms of bytes (valied only if msgType is data)

	//these are the values that are not actually sent over the air but are for modeling channel nose and signal decay
	double rssi;//set the signal strength value of the packets at the receiver
};

struct EventInfo {
	short nodeId_; //the id of the node for which this event occured, 
	Packet packet_;///the container that has information about the pkt length/destination etc. 
	eventType eType_; /// event Type... reception or transmission.
	timeTicks timeLeft_;//assigned with the time left till the event expiry	
};

/***
* The Classes used for implementing the concept of node and node clock
*
***/


///the Event Timer class that implements the discrete event simulation 

class Event: public TimerCallback{
	
	private:
		EventInfo eInfo_;	

	public:
		Event(EventInfo eInfo) : eInfo_(eInfo){};
		~Event(){};
		int Expire();

};

/// The implementation of the node clock that handles skew and offset
class cNodeClock{
	public:
	//by default assume a perfectly calibrated clock
		cNodeClock(float skew=1.0,float offset =0.0): skew_(skew), offset_(offset){};
		~cNodeClock(){};
	///this inline functione will return the erroneous time at the local node. 
		timeTicks getLocalTime(timeTicks realTime) { 
			float lt= (realTime*skew_)+offset_;
			timeTicks retVal = (timeTicks)granularityConversion(lt);
			return retVal; 
		};

private:
	float  skew_;
	float offset_;


};

#endif
