/*
 *lineReg.cpp
 * Author: Affan Syed <asyed@usc.edu>
 *  Definition of the class for doing linear regression on input data points
 *   
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */
#include <math.h>
#include <float.h>
#include "lineReg.h"

//this function is used to feed/update the linear regression engine
void cLinearRegression::AddPoint(const DataPoint & dataPoint){ 
	
  /*n++;//increment the count of the data point fed to the linear regression engine
	sumX += array[n-1][0];
	sumY += array[n-1][1];
	sumXY += array[n-1][0] * array[n-1][1];

	sumXsquared +=array[n-1][0]*array[n-1][0];
	printf("X=%2f, Y=%2f \n",array[n-1][0], array[n-1][1]);
  */
	//// Modify all the sums according to the current input.
	//printf("X=%2f, Y=%2f \n",dataPoint.localTime, dataPoint.offset);
	sumX += dataPoint.localTime;
	sumY += dataPoint.offset; ///offset is global time-local time
	//sumY += dataPoint.rxTime; //ALT WAY
	sumXsquared += dataPoint.localTime* dataPoint.localTime;
	sumYsquared += dataPoint.offset* dataPoint.offset;
	sumXY += dataPoint.localTime*dataPoint.offset;
	//sumXY += dataPoint.localTime*dataPoint.rxTime;//ALT way

}

bool  cLinearRegression::Calculate(){
	bool retFlag= false;
	double num,denom;
	 
	if (HaveData())
	  {
//		printf("sumX= %f, sumY=%f, sumXSq=%f, sumXY=%f \n", sumX, sumY, sumXsquared, sumXY);
	//	printf(" Number of beacons = %d\n", n);
		  if (fabs( double(n) * sumXsquared - sumX * sumX) 
			  > DBL_EPSILON) //sanity check to prevent divide by zero!
		  {
//		  fprintf(stderr,"&&&&&Got into Regression Calc!!!&&&&&\n");
			  num =( double(n) * sumXY - sumY * sumX);
			  denom =  ( double(n) * sumXsquared - sumX * sumX);
			  a =num/denom;
			  
			  b = (sumY - a * sumX) / double(n);
		//	  printf("numerator =%f, denom = %f \n",num, denom);
			//printf("A=%f, B= %f --- Size of= %d", a,b, sizeof(a));
			  double sx = a * (sumXY - sumX * sumY / double(n));
			  double sy2 = sumYsquared - sumY * sumY / double(n);
			  double sy = sy2 - sx;
			  coefD = sx / sy2;
			  coefC = sqrt(coefD);
			  stdError = sqrt(sy / double(n - 2));
			  retFlag = true;
//		  	fprintf(stderr,"&&&&&Linear regression Result a=%f, b=%f  &&&&&\n",a,b);
		  }
		  else
			  
		  {
			  a = b = coefD = coefC = stdError = 0.0;			  
		  }
	  }

	
	return retFlag;

}



