
/*
 * Author: Affan Syed <asyed@usc.edu>
 *  lineReg.h: Declaration of class for linear regression
 *   
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */
#ifndef _LINREG_H_
#define _LINREG_H_

#include "defines.h"

//this data structure holds the data to be used for linear regression
struct DataPoint{
	double  localTime;
	double rxTime;
	double  offset;//offset = gTime-LocalTime 
	//BUGFIXED: Since offset *can* be negative ... we have to allow for that
	bool isValid; ///maintain the validity of a datapoint - invalid if beacon not received

	DataPoint(): localTime(0), offset(0), isValid(false){};
};

// a linear regression analysis class
class cLinearRegression
{
    

    public:
		///default constructor... init every varaible to zero
        cLinearRegression(){n=0;
		sumX=sumXsquared=sumXY=sumY=sumYsquared=a=b=0.0;};

		void resetAll() {n=0;
		sumX=sumXsquared=sumXY=sumY=sumYsquared=a=b=0.0;};
        void AddPoint(const DataPoint& dataPoint); 
	bool Calculate();// calculate coefficients- call this after you have input all the parameters 
		   //return false if not enough data points available to make decision
	void UpdateOffset(double offset) { b=offset;};

        //Do we have enough data to perform calculations (min 3)
        int HaveData() const { return (n > 2 ? 1 : 0); }
        
	 double GetSlope() const { return a; }
	double GetOffset() const { return b; }

        double GetCoefDeterm() const  { return coefD; }
        double GetCoefCorrel() const { return coefC; }
        double GetStdErrorEst() const { return stdError; }

	double EstimateY(double x) const  { return (a* x+b); }

    protected:
        long n;             // number of data points input
        double sumX, sumY;  // sums of x and y
        double sumXsquared, // sum of x squares
               sumYsquared; // sum y squares
        double sumXY;       // sum of x*y

        double a, b;        // coefficients of f(x) = a*x +b
        double coefD,       // coefficient of determination
               coefC,       // coefficient of correlation
               stdError;    // standard error of estimate

       
};





#endif 	/* end linRege.h*/

