/*
 * Author: Affan Syed <asyed@usc.edu>
 *   MAC.h: This defines the container class for the MAC. specific incarnations need to be derived from this base class
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */

#ifndef _MAC_H_
#define _MAC_H_


#include "defines.h"
#include "phy.h"

struct rxPktInfo{
bool urDestination;//tells if the node is destination of the rx packet
timeTicks endOfPktReception;//informs when the packet will end based on its length
};

class cMAC {

public:
	uint_32 m_pktsSent;//here we have statistical variables
	uint_32 m_pktsDropped;//application pkts blocked b/c of exisiting contention. 	
	uint_32 m_pktsCollided;//keeps track of data packets that collided (at a receiver!!) only due data overlap
	uint_32 m_pktsCollidedByTone;//# of data pkts that corrupted by tone
	uint_32 m_pktsReceived;//pkts received as a dest
	uint_32 m_pktsLost;///the tx'ed pkts lost by data collision - but NOT broadcast pkts lost
	uint_32 m_pktsLostByTone;///the tx'ed pkts lost by tone collision- but NOT broadcast pkts lost
	double 	MAX_RANGE;			
	double MAX_PROP_DELAY;		
	double   	MAX_RTT;		
	uint_16 numNodes;//required for local scheduling of transmission slots in TDMA---total # of nodes in the simulation
	
	double totalLatency;//TAG:latencymeasure
	timeTicks startBufferTime;//TAG:latencymeasure
				
	short m_myId;
	nodeState m_macState;
	bool Idle;//this gives the binary state of whether a MAC already has a data pkt to send or not
	macParameter m_macParam;
	Packet m_currPacket;//this struct will contain the current packet that needs to be transmitted (for which the MAC is contending)
	rxPktInfo rxPkt;//the pkt that we are currently receiving. (we sync to its length and need to turn off our AM)
	FILE* outFile;
	handle hdlTimeOut;///this keeps a handle on the TO event that might need to be removed
	
	/**
	The next two classes keep track of power consumed 
	by each of our transducers

	***/
	cAcousticModemPower amPower;
	cWakeupPower wuPower;
	

	cPhy underwaterPhy;

	cMAC(short Id,FILE* fptr, macParameter macParam):m_myId(Id),outFile(fptr),
			m_macParam(macParam),underwaterPhy(Id,&amPower,&wuPower, macParam.toneCancelProb){ 
		MAX_RANGE= macParam.maxRange;
		MAX_PROP_DELAY= ((double)MAX_RANGE/SNDSPEED_MPS );//in seconds
		MAX_RTT=2*MAX_PROP_DELAY;//in seconds
		m_pktsSent=m_pktsReceived=m_pktsDropped=m_pktsCollided=m_pktsLost=m_pktsLostByTone=0;
		totalLatency=0;
		startBufferTime=0;
		m_macState =e_mac_idle;
		Idle=true;

	};
	virtual ~cMAC(){};

	void setNodeCount(int count){numNodes=count;};
	void printStats();
	//define  pure virtual functions that will be implemented by the derived classes
	
 	virtual bool isIdle()=0;
	virtual bool resetMAC()=0;
		
	virtual bool macSend(Packet pkt) =0;//this will return false if mac is currently busy sending last packet

	
	//All these are to control the input from the simulation core loop
/*********
phyinterface UPDATE	

virtual bool macToneRxStart(EventInfo& eventInformation) =0; 
virtual  bool macToneTxEnd(EventInfo& eventInformation)=0;

*****/
	virtual  bool macToneDetected(EventInfo& eventInformation)=0;

	
	virtual bool macDataRxStart(EventInfo& eventInformation, channelState initState) =0;
	
	virtual  bool macDataRxEnd(EventInfo& eventInformation, channelState initState)=0;
	virtual  bool macDataTxEnd(EventInfo& eventInformation, channelState initState)=0;

//this has MAC dependant interpretation. Basically gives a TO if the MAC has set up one
	virtual  bool macTimerExpire(EventInfo& eventInformation)=0;
	
		


};
#endif
