/*
 * Author: Affan Syed <asyed@usc.edu>
 *   phy.cpp: The physical layer abstraction 
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */

 
#include "phy.h"

extern	cVTime globalClock;///Our global clock!
extern	Timers gTimerManager;
extern 	cChannel gChannel;


/****
COMMENT::
AFFAN: 12/25/2007: 
I am not entirely happy with the way i am doing power transitions
here. There is an unnecessary overlap between the MAC and phy layer
where I am making the phy layer make additional transition to the wu (wake up)
or acoustic modem (am) hardware-- something which works for T-Lohi
but starts failing when I have other MAC (like TDMA and CSMA). 

I might need to redo this and that would imply ensuring that the correct transitions
are being made in my T-Lohi , which would take some thinking. 

The other solution is to make reverse transitions in the new MAC's... always a bad idea
but easier in the short run. 

PS: A more carefull look inidicates that the phy is quite generic for the am, 
 but the behavior of the wu transition is hardcoded to reflect that of T-Lohi... 
 so we might be okay if we ignore the power values associated with wu!?



POWER CONSUMPTION TRANSITIONS:
--------------------------------
The phy layer makes the transition from tx'ing to rx'ing/or idle mode of a
radio device by itself, provided that the device has not been turned off (by an upper
layer). 

So the following rules are in place:

1)If WU transmission ends, and there is data on the AM channel
 then if AM is not off, set AM and WU to Rx state (with assumption that
 if data is present and WU is on, it takes same amount of enerfy as if
 it was rxing tone). If no data but tone present, WU->Rx, if neither present
 turn both data (if not off) and WU to on/listen.

 2)If AM transmission ends, then if data is present
  change am and WU (if not off) to off otherwise to on/listen. 
  If tone is present (and no data) turn the WU to rx other wise to On. 

  3)Every time a tone reception ends, we check for the same stuff as in 1,
  except that we also check if out WU tone is not off. 
  4) Same as 3) for data except that we do the stuff in 2)

  5)On the start of tone or data transmission we change the states of 
   the devices to tx without any care about their previous state (as it should be)

   6) On start of a tone, we decide based on what state the channel is in, to turn on 
    WU or not (if it is not off!). So if RTS is in a data state (Present, Overlap, tone overlap)
    then we donot alter the WU dev state. Only if the RTS is in floor state to we change to WU->rx,
    as in other cases it should already be in rx if it wasnt off. 
    7) for and RTE, the only time we change WU is when we go to the floor state, and that too we go
    to On/listen state if it is not off. 

    8) for RDS, we turn on our AM to rx, if it is not in off state. or not already turned on !

    9) for RDE, we change to AM-> rx if there is still other data on the channel
     or otherwise we change to AM-> on (Both provided AM is not off) .
     Based on the state that it occurs we need to be careful to change the WU to rx
     or tx. So e.g. if the RDE transits to tone overlap then we check if WU is not already transmitting
     in which case we dont change the state to rx, otherwise we do 
     (ofcourse when WU is not off)

***/



/********
This is the TTS event for the phy layer.
No matter what the state of the channel is, a tone will always be
put on the channel. A corresponding end of tone event will also be generated. 
*****/

channelState cPhy::PhySendTone(){
	Packet pktTx;
	TimerCallback* tcbPtr;///pointer to insert timers for each node
	EventInfo eventInformation;
	timeTicks tempTCCT;
	channelState initState;

	initState= channelStateAtNode;
/***** 
X1 as stated in the state machine diagram 
*****/

//update the flag mentioning that we are transmitting
	toneTxing=true;
///first always update the TCCT if it needs to be extended
	eventInformation.timeLeft_= (timeTicks)(TONE_DURATION*SEC_TO_TIMETICK);
	tempTCCT=globalClock.globalTime+eventInformation.timeLeft_;
	
	if( TCCT < tempTCCT) //i.e TCCT needs to be extended
		TCCT= tempTCCT;
	//else the tone channel is busy beyond this event, dont decrease the TCCT value, ever!!

	
///also set up the TTE event right now
	eventInformation.eType_=e_tone_tx_end;
	eventInformation.nodeId_=m_myId;
	tcbPtr = new Event(eventInformation);

       hdlToneTxEnd = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed


//another thing we always do with tone Tx is to
//check if the TD event was set, if it was, it should 
//be interrrupted and the event removed from the queue
if(TD.isSet==true){
	gTimerManager.RemoveTimer(TD.eventHandle);
	TD.isSet=false;
}
	
/**************************************************************
						X1 ends
**************************************************************/
/******
Now update the state of the 
channel based on the state 
machine
*******/
//for avg slot test - in case of UT-Lohi
	//cout<<"start "<<globalClock.globalTime/SEC_TO_TIMETICK<<endl;

switch(channelStateAtNode){

		case e_floor:{
						
			channelStateAtNode = e_tonePresent;

			OUTCON(m_myId,  " cPhy::TTS->floor", "",3);	
			break;}
		case e_dataPresent:{
			//since X1 has already been performed, only the state transition needs to be done
			channelStateAtNode = e_dataToneOverlap;
			break;
			}
		case e_dataOverlap:{
			///a data overlap trumps tone overlap
			//i.e. if data are overlapped the corresponding
			// data will never be good, while if tone data
			//were to overlap it depends on whether tone/data are orthognal.
			//hence no state change at all.  hence here we stay in the higher corruption 
			//state of dataOverlap, and X1 has already been performed
			channelStateAtNode = e_dataOverlap;
			OUTCON(m_myId,  " cPhy::TTS->dataOverlap", "",3);	
			break;
			}
		case e_tonePresent:{

			//this will only cause a tone tone overlap: X1 already performed

			channelStateAtNode=e_toneOverlap;
			//if no other tone is ever received in this new state the 
			// RTE will be ignored and TTE will reset everything
			OUTCON(m_myId,  " cPhy::TTS->tonePresent", "",3);	
			break;}
		case e_toneOverlap:{
			//doesnt change the state at all, 
			channelStateAtNode=e_toneOverlap;			
			//also perfrom X1, already done so
			OUTCON(m_myId,  " cPhy::TTS->toneOverlap", "",3);				
			break;
			}
		case e_dataToneOverlap:{
			//doesnt change the state at all, 
			channelStateAtNode=e_dataToneOverlap;			
			//also perfrom X1, already done so
			OUTCON(m_myId,  " cPhy::TTS->dataToneOverlap", "",3);	
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::TTS->UNKNOWNSTATE", "",3);	
			break;

	}


/**********/

///Now lets make the simulator send the tone to all nodes
	pktTx.msgType_=e_msg_tone;
	pktTx.sourceId_ =m_myId;///this is so that the channel can set up the correct receive times based on distance from this node to other nodes
	//all Other parts of the pkt are not useful for the tone case.. no need to initialize them. 

//set the WU state to transmitting
	wuPower->changeDevState(e_wakeup_tx, globalClock.globalTime);
	 gChannel.sendTone(pktTx);



	 
	return initState;
}

/********
Handle the start of tone energy on the physical channel.
This is the RTS event for the phy layer
*****/
channelState cPhy::PhyToneRxStart(EventInfo eventInformation){
timeTicks tempTCCT;
TimerCallback* tcbPtr;///pointer to insert timers for each node
channelState initState;


/*** The following 2 lines of code implement tone drop capability of our simulator ****/
if(RAND_UNIFORME < m_toneCancelProb){
	return channelStateAtNode;
}//otherwise we simply continue as normal


/****
first always update the TCCT
and make sure a corresponding RTE event is generated
****/
eventInformation.timeLeft_= (timeTicks)(TONE_DURATION*SEC_TO_TIMETICK);
tempTCCT= globalClock.globalTime+eventInformation.timeLeft_;
if( TCCT < tempTCCT) //i.e TCCT needs to be extended
		TCCT= tempTCCT;
//else the tone channel is busy beyond this event, dont decrease the TCCT value, ever!!

	

//in any case set up an event that will signal the end of eneergy on the tone channel
eventInformation.eType_=e_tone_rx_end;
eventInformation.nodeId_=m_myId;
tcbPtr = new Event(eventInformation);

hdlToneEnd =gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed


initState=channelStateAtNode;
//////////////////////////////////////

/******
Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){

		case e_floor:{
			//doing Z

			//set up a tone detection event
			eventInformation.timeLeft_=(timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK);
			eventInformation.eType_=e_mac_toneDetected;
			eventInformation.nodeId_=m_myId;
			tcbPtr = new Event(eventInformation);
			TD.eventHandle = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle 
			TD.isSet= true;
			
			//state transit
			channelStateAtNode = e_tonePresent;

			if(wuPower->currentState ==e_wakeup_on){
				wuPower->changeDevState(e_wakeup_rx, globalClock.globalTime);
			}			
			else{
				
				OUTCON(m_myId,  " cPhy::RTS->floor", "WU not in listen mode! POWER CONTROL ERR",2);	
			}
			OUTCON(m_myId,  " cPhy::RTS->floor", "",3);	
			break;}
		case e_dataPresent:{
			//nothing else other than extending the TCCT and change our state
			//TCCT already extended, just changing our state
			channelStateAtNode = e_dataToneOverlap;
			
			OUTCON(m_myId,  " cPhy::RTS->dataPresent", "",3);	
			break;
			}
		case e_dataOverlap:{
			//here only TCCT needs to be extended
			//the state remains the same (data overlap trumps)

			channelStateAtNode= e_dataOverlap;//no change
			
			OUTCON(m_myId,  " cPhy::RTS->dataOverlap", "",3);	
			break;
			}
		case e_tonePresent:{
			//only update TCCT (done)
			//and change state (below)
			channelStateAtNode= e_toneOverlap;
			OUTCON(m_myId,  " cPhy::RTS->tonePresent", "",3);				 
			break;}
		case e_toneOverlap:{
			//stay where u are extend TCCT if needed (done)
			channelStateAtNode= e_toneOverlap;//no change			
			OUTCON(m_myId,  " cPhy::RTS->toneOverlap", "",3);				 
			break;
			}
		case e_dataToneOverlap:{
			//stay where u are extend TCCT if needed (done)
			channelStateAtNode= e_dataToneOverlap;//no change			
			OUTCON(m_myId,  " cPhy::RTS->dataToneOverlap", "",3);				 
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::TTS->UNKNOWNSTATE", "",3);	
			break;

	}

/**********/



return initState;//that this was succesful, but make sure that the calle return false to the event loop so that the pointers are deleted
	
}



/******

This is the TDS event for the phy layer state machine 

****/
channelState cPhy::PhySendData(Packet& pktTx){
	
	TimerCallback* tcbPtr;///pointer to insert timers for each node
	EventInfo eventInformation;
	timeTicks  tempDCCT;
	 channelState initState= channelStateAtNode;


/****
first always update the DCCT
and make sure a corresponding TDE event is generated
****/
eventInformation.timeLeft_=  (timeTicks) ((pktTx.length_/CHANNEL_DATA_RATE)*SEC_TO_TIMETICK);//the data duration left in time tick

tempDCCT=globalClock.globalTime+eventInformation.timeLeft_;

if( DCCT < tempDCCT) //i.e DCCT needs to be extended
	DCCT= tempDCCT;
//else the data channel is busy beyond this event, dont decrease its value, ever!!

///also set up the TTE event right now
eventInformation.eType_=e_data_tx_end;
eventInformation.nodeId_=m_myId;
tcbPtr = new Event(eventInformation);

hdlDataTxEnd = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed


//Doing X2 every time
//i.e. if a TD event was set, 
//remove it 

if(TD.isSet==true){
	gTimerManager.RemoveTimer(TD.eventHandle);
	TD.isSet=false;
}

///////////////////////////////////////////


//for avg slot test
	//cout<<"end "<<globalClock.globalTime/SEC_TO_TIMETICK<<endl;


/******
Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){

		case e_floor:{
			//extending the DCCT(done)
			//do X2 (done)
			
			//change state (below)
			channelStateAtNode = e_dataPresent;

			OUTCON(m_myId,  " cPhy::TDS->floor", "",3);	
			break;}
		case e_dataPresent:{
			//extend DCCT (done)
			//do X2 (done)
			//change state (below)
			channelStateAtNode = e_dataOverlap;
			
			OUTCON(m_myId,  " cPhy::TDS->dataPresent", "***** This should never happen *****",2);
			//the reason none of the error cases on TDS should ever happen is that
			// a MAC should be intelligent enough to not transmit a data if there is some 
			//data already being received (or present)
			break;
			}
		case e_dataOverlap:{
			//extend DCCT (done)

			channelStateAtNode= e_dataOverlap;//no change
			OUTCON(m_myId,  " cPhy::TDS->dataOverlap", "***** This should never happen *****",2);								
			break;
			}
		case e_tonePresent:{
			//extend DCCT (done)
			//do X2(done)
			//change state 
			channelStateAtNode= e_dataToneOverlap;
			OUTCON(m_myId,  " cPhy::TDS->tonePresent", "",3);				
			break;}
		case e_toneOverlap:{
			//extend DCCT (done)
			//do X2(done)
			//change state 
			channelStateAtNode= e_dataToneOverlap;
			OUTCON(m_myId,  " cPhy::TDS->toneOverlap", "",3);				
			break;
			}
		case e_dataToneOverlap:{
			//extend DCCT (done)
			//do X2(done)
			//change state 
			//(note that since there was atleast
			//one data already present we now have both
			//data data and data tone overlap
			//based on the data-data TRUMPS data-tone
			//we change state)

			channelStateAtNode= e_dataOverlap;
			OUTCON(m_myId,  " cPhy::TDS->dataToneOverlap", "***** This should never happen *****",2);							
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::TDS->UNKNOWNSTATE", "",3);
			break;

	}



/**********/

///Now send the actualy pkt to simulator so that it can be transmitted to all nodes
pktTx.msgType_ =e_msg_data;//make sure that it gets interpreted as such
pktTx.sourceId_ =m_myId;///this is so that the channel can set up the correct receive times based on distance from this node to other nodes
	
//change its state to tx- generically.. our data transfer mechanism, for T-Lohi,  will change this as a hack!!!
amPower->changeDevState(e_am_tx,globalClock.globalTime);

gChannel.sendData(pktTx);

return initState;
}


/***
This is the RDS event
***/
channelState cPhy::PhyDataRxStart(EventInfo eventInformation){
TimerCallback* tcbPtr;///pointer to insert timers for each node
timeTicks tempDCCT;
channelState initState = channelStateAtNode;


/****
first always update the DCCT
and make sure a corresponding RDE event is generated
****/
eventInformation.timeLeft_= (timeTicks)((eventInformation.packet_.length_ /CHANNEL_DATA_RATE)*SEC_TO_TIMETICK);
tempDCCT= globalClock.globalTime+eventInformation.timeLeft_;
//HACKKK
if(DCCT==tempDCCT){
	DCCT++;//just so that two or more overlaping data receptions occur 
	//one of them clears one time tick later
	eventInformation.timeLeft_++;//also make this particular event trigger later so that it can
	//end the data reception correctly. 

}
if( DCCT < tempDCCT) //i.e DCCT needs to be extended
	DCCT= tempDCCT;

//else the data channel is busy beyond this event, dont decrease its value, ever!!


//in any case set up an event that will signal the end of eneergy on the tone channel
eventInformation.eType_=e_data_rx_end;
eventInformation.nodeId_=m_myId;
tcbPtr = new Event(eventInformation);

hdlDataTxEnd =gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle so that you can remove it if needed

//Doing X2 every time
//i.e. if a TD event was set, 
//remove it 

if(TD.isSet==true){
	gTimerManager.RemoveTimer(TD.eventHandle);
	TD.isSet=false;
}

//////////////////////////////////////

/******
Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){//NOTE: THE RDS EVENT behave EXACTLY the same as TDS event, always

		
		case e_floor:{
			//extending the DCCT(done)
			//do X2 (done)
			
			//change state (below)
			channelStateAtNode = e_dataPresent;


			if(wuPower->currentState!=e_wakeup_off)//if the wu is not off it will transit to a receive state as data energy will trigger its reception circuit
				wuPower->changeDevState(e_wakeup_rx,globalClock.globalTime);//NOTE:: This might cause wu to consume energy even in MAC's 
				//where we dont assume it is present.. not a generic thing to do.. but should be kept in mind while giving energy figures
				
			//also do the same for am (this is a generic phy layer.. hence we are doing so. for lohi the mac will always keep am off)
			if(amPower->currentState!=e_am_off){
					amPower->changeDevState(e_am_rx,globalClock.globalTime);
				}


			OUTCON(m_myId,  " cPhy::RDS->floor", "",3);	
			break;}
		case e_dataPresent:{
			//extend DCCT (done)
			//do X2 (done)
			//change state (below)
			channelStateAtNode = e_dataOverlap;
			
			OUTCON(m_myId,  " cPhy::RDS->dataPresent", "",3);				
			break;
			}
		case e_dataOverlap:{
			//extend DCCT (done)

			channelStateAtNode= e_dataOverlap;//no change
			OUTCON(m_myId,  " cPhy::RDS->dataOverlap", "",3);				
			break;
			}
		case e_tonePresent:{
			//extend DCCT (done)
			//do X2(done)
			//change state 
			channelStateAtNode= e_dataToneOverlap;
			OUTCON(m_myId,  " cPhy::RDS->tonePresent", "",3);				
			break;}
		case e_toneOverlap:{
			//extend DCCT (done)
			//do X2(done)
			//change state 
			channelStateAtNode= e_dataToneOverlap;
			OUTCON(m_myId,  " cPhy::RDS->toneOverlap", "",3);				
			break;
			}
		case e_dataToneOverlap:{
			//extend DCCT (done)
			//do X2(done)
			//change state 
			//(note that since there was atleast
			//one data already present we now have both
			//data data and data tone overlap
			//based on the data-data TRUMPS data-tone
			//we change state)

			channelStateAtNode= e_dataOverlap;
			OUTCON(m_myId,  " cPhy::RDS->dataToneOverlap", " >> dataOverlap",3);				
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::RDS->UNKNOWNSTATE", "",3);
			break;

	}



/**********/

return initState;//that this was succesful, but make sure that the calle return false to the event loop so that the pointers are deleted


}


/**
Signals that the tone rx end event has been received at MAC. 
This is the RTE even for the phy layer
**/

channelState cPhy::PhyToneRxEnds(EventInfo eventInformation){

channelState initState = channelStateAtNode;


/******
Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){

		case e_floor:{
			OUTCON(m_myId,  " cPhy::RTE->floor", "***STATE M/C ERROR ***",3);
			break;}
		case e_dataPresent:{
			OUTCON(m_myId,  " cPhy::RTE->dataPresent", "***STATE M/C ERROR ***",3);			
			break;
			}
		case e_dataOverlap:{

			//do nothing, stay in same state
			channelStateAtNode= e_dataOverlap;//no change
			
			OUTCON(m_myId,  " cPhy::RTE->dataOverlap", "",3);			
			break;
			}
		case e_tonePresent:{
			//this should be corroborated with a tone detect event ALWAYS
			//change state to floor as all 
			channelStateAtNode = e_floor;

			//also change state of wakeup to on/listen/idle
			if(wuPower->currentState!=e_wakeup_off)
				wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);
	
			OUTCON(m_myId,  " cPhy::RTE->tonePresent", "",3);
			
			break;}
		case e_toneOverlap:{
			//over here we have two situations
			if(TCCT > globalClock.globalTime){
			// 1) there is still some tone energy present on the channel
			//in this case we stay in this state and simply wait for it to end
			channelStateAtNode = e_toneOverlap;

			OUTCON(m_myId,  " cPhy::RTE->toneOverlap", ">>toneOverlap",3);

			}
			else{
			// 2) the end of this tone corresponds to an ends 
			// of any tone at this node

			//first we delete any TD event that might have been set but hasnt occured 
			//(only due to transition from data overlap or tonedata overlap)
			/******* CAUTION *******

			There is a single case where removing TD event can 
			result in a wrong conclusion. i.e. when 2 tone
			events are received at the same time, causing a TD event
			to be setup due to the first. But depending on what order the events are
			executed, it might occur that the RTE event for both happen 
			before the TD event occure (all set for THE SAME TIME),
			and we delete the TD event even though it shoudl have been received. 
			************************/
				if(TD.isSet==true){//perform modified X1

					//:: POTENTIALLY BUGGGY CODE:://
					//first we check the time Left for the TD to happen
					timeTicks tLeft = gTimerManager.getTimerTimeLeft(TD.eventHandle);
					if(tLeft != 0 && tLeft != MAX_TIMER_VALUE){ //this is to handle the caution above
						gTimerManager.RemoveTimer(TD.eventHandle);
						TD.isSet= false;
						}					
				}
			
			channelStateAtNode =e_floor;
			
			//also change state of wakeup to on/listen/idle
			if(wuPower->currentState!=e_wakeup_off)
				wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);
			
			OUTCON(m_myId,  " cPhy::RTE->toneOverlap", ">>floor",3);

			}
			break;
			}
		case e_dataToneOverlap:{
			//nothing to do
			channelStateAtNode= e_dataToneOverlap;//do nothing
			OUTCON(m_myId,  " cPhy::RTE->dataToneOverlap", "",3);
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::RTE->UNKNOWNSTATE", "",3);
			break;

	}

/**********/

return initState;	
}


//the TTE event
channelState cPhy::PhyToneTxEnds(EventInfo eventInformation){
channelState initState= channelStateAtNode;
TimerCallback* tcbPtr;///pointer to insert timers for each node

//make sure that you turn off the Ttx'ing flag
toneTxing = false;


/******

Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){

		case e_floor:{
				OUTCON(m_myId,  " cPhy::TTE->floor", "***STATE M/C ERROR ***",3);
			break;}
		case e_dataPresent:{
				OUTCON(m_myId,  " cPhy::TTE->dataPresent", "***STATE M/C ERROR ***",3);			
			break;
			}
		case e_dataOverlap:{
			//set tone txing to false (done)

			channelStateAtNode=e_dataOverlap;//no change
			
			OUTCON(m_myId,  " cPhy::TTE->dataOverlap", "",3);			
			break;
			}
		case e_tonePresent:{
			//set tone txing to false (done)

			//transit to floor state as no tone should be left after this
			channelStateAtNode= e_floor;
			OUTCON(m_myId,  " cPhy::TTE->tonePresent", "",3);			
			break;}
		case e_toneOverlap:{
			//set tone txing to false (done)

			//now check if there is still some tone energy remaining (from receiving overlaps)
			if(TCCT> globalClock.globalTime){//i.e. there is !!

				//Do Z (no need for Y as we - b definition- turned the tTx flag to false)

				//doing Z

				//set up a tone detection event
				eventInformation.timeLeft_=(timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK);
				eventInformation.eType_=e_mac_toneDetected;
				eventInformation.nodeId_=m_myId;
				tcbPtr = new Event(eventInformation);
				TD.eventHandle = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle 
				TD.isSet=true;


				channelStateAtNode= e_toneOverlap;
				OUTCON(m_myId,  " cPhy::TTE->tonePresent", ">>toneOverlap",3);			
				}
			else{
			
			//do X2
				if(TD.isSet==true){
					TD.isSet=false;
					gTimerManager.RemoveTimer(TD.eventHandle);
				}

				//in this case there is no energy left on the tone channel- hence we are clear
				channelStateAtNode= e_floor;
				OUTCON(m_myId,  " cPhy::TTE->tonePresent", ">>floor",3);						
			}
		
			break;
			}
		case e_dataToneOverlap:{
			//set tone txing to false (done)
			channelStateAtNode=e_dataToneOverlap;//no change
			OUTCON(m_myId,  " cPhy::TTE->dataToneOVerlap", "",3);									
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::TTE->UNKNOWNSTATE", "",3);
			break;

	}

/**********/
//also change the WU state so that now it is in the on or rx state (its bound to be in tx state right now!!!)
wakeUpOn( globalClock.globalTime);
//also turn on the AM 
if(amPower->currentState !=e_am_off)
	amOn(globalClock.globalTime);

	return initState;
}

/**
Signals that the data rx end event has been received 
Some data sources energy/oacket has just ended
**/

channelState cPhy::PhyDataRxEnds(EventInfo eventInformation){

channelState returnState=channelStateAtNode;//initialize with the current state
TimerCallback* tcbPtr;///pointer to insert timers for each node

/******
Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){

		
		case e_dataPresent:{//this is great.. a correct data reception
			//do X2 (i am being anal, but there should never be a TD occuring
			if(TD.isSet){
//			cout<<"Tone detect event set.. canceling, but not sure why!!"<<endl;
			TD.isSet=false;
			gTimerManager.RemoveTimer(TD.eventHandle);
			}
			//channel is totaly clear now
			channelStateAtNode= e_floor;
			//also change state of wakeup to on/listen/idle
			if(wuPower->currentState!=e_wakeup_off)
				wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);

			if(amPower->currentState!=e_am_off)// change AM to listen now if it was not already off (as it should be except for @ rx)
				amOn(globalClock.globalTime);//the modem has to be explicity turned off	
				
					
			OUTCON(m_myId,  " cPhy::RDE->dataPresent", "",3);
			break;
			}
		case e_dataOverlap:{
			//this case has many different scenarios
			timeTicks CT= globalClock.globalTime;
			//first we check if there is any other data present after
			//this data ends
			if(DCCT > CT){//yes
				//in this case simply return to the orignal state
				channelStateAtNode= e_dataOverlap;
				OUTCON(m_myId,  " cPhy::RDE->dataOverlap", ">>dataOverlap",3);						
			}
			else{//all data energy is absent
			//in this case change the return state to show that all data has finished on the
			//data channel, but this was still a corrupted packet
			returnState = e_dataOverlapEnded;
			
				if(amPower->currentState!=e_am_off)// change it to listen now if it was not already off (as it should be except for @ rx)
					amOn(globalClock.globalTime);//the modem has to be explicity turned off	
					
			//now check if any tone energy remains
				if(TCCT> CT){//yes there is some tone energy left
				
				//do Y
				
				//first check if the energy isnt due to our own tone transmission
					if(toneTxing==false){//In this case we have to setup a new TD event
						eventInformation.timeLeft_=(timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK);
						eventInformation.eType_=e_mac_toneDetected;
						eventInformation.nodeId_=m_myId;
						tcbPtr = new Event(eventInformation);
						TD.eventHandle = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle 
						TD.isSet=true;
					}
				//transit to the tone OVerlap state (eventhough there might not be any overlap in reality)
				//as it doesnt matter in the transition
				channelStateAtNode= e_toneOverlap;

				//also change state of wakeup to listen if its not transmitting or off, otherwise keep in same state
				if(wuPower->currentState!=e_wakeup_off && wuPower->currentState!=e_wakeup_tx)
					wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);

				OUTCON(m_myId,  " cPhy::RDE->dataOverlap", ">>toneOverlap",3);						
				}
				else{//no type of energy remians on the channel
					//do X2
					if(TD.isSet){			
						TD.isSet=false;
						gTimerManager.RemoveTimer(TD.eventHandle);
					}
					channelStateAtNode=e_floor;
					
					//also change state of wakeup to on/listen/idle
					if(wuPower->currentState!=e_wakeup_off)
						wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);					
					
					OUTCON(m_myId,  " cPhy::RDE->dataOverlap", ">>floor",3);						
				}


			}

			break;
			}
		case e_dataToneOverlap:{
			//note we are in dataToneOverlap state iff we have just a single data present. if more than one data
			//present, the dataDataOverlap state holds trumps. 
			
			if(amPower->currentState!=e_am_off)// change it to listen now if it was not already off (as it should be except for @ rx)
					amOn(globalClock.globalTime);//the modem has to be explicity turned off	
				

			//check the state of tone energy after this event
			if(TCCT > globalClock.globalTime){//there is still tone energy present
				//do Y
				//first check if the energy isnt due to our own tone transmission
				if(toneTxing==false){//In this case we have to setup a new TD event
					eventInformation.timeLeft_=(timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK);
					eventInformation.eType_=e_mac_toneDetected;
					eventInformation.nodeId_=m_myId;
					tcbPtr = new Event(eventInformation);
					TD.eventHandle = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle 
					TD.isSet=true;
				}
				//transit to the tone OVerlap state (eventhough there might not be any overlap in reality)
				//as it doesnt matter in the transition
				channelStateAtNode= e_toneOverlap;
				
				//also change state of wakeup to listen if its not transmitting or off, otherwise keep in same state
				if(wuPower->currentState!=e_wakeup_off && wuPower->currentState!=e_wakeup_tx)
					wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);

				
				
				OUTCON(m_myId,  " cPhy::RDE->dataToneOverlap", ">>toneOverlap",3);						
			}
			else{//no type of energy remians on the channel
				//do X2
				if(TD.isSet){			
					TD.isSet=false;
					gTimerManager.RemoveTimer(TD.eventHandle);
				}
				channelStateAtNode=e_floor;
				//also change state of wakeup to listen if its not off
				if(wuPower->currentState!=e_wakeup_off )
					wuPower->changeDevState(e_wakeup_on, globalClock.globalTime);
							
					
				OUTCON(m_myId,  " cPhy::RDE->dataToneOverlap", ">>floor",3);						
			}
				

			break;
			}
		case e_floor:{
			OUTCON(m_myId,  " cPhy::RDE->floor", "***STATE M/C ERROR ***",3);
			break;
			}
		case e_tonePresent:{
			OUTCON(m_myId,  " cPhy::RDE->tonePresent", "***STATE M/C ERROR ***",3);
			break;
			}
		case e_toneOverlap:{
			OUTCON(m_myId,  " cPhy::RDE->toneOverlap", "***STATE M/C ERROR ***",3);			
			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::RDE->UNKNOWNSTATE", "",3);
			break;

	}



/**********/
return returnState;
}

//the TDE event
channelState cPhy::PhyDataTxEnds(EventInfo eventInformation){

channelState returnState=channelStateAtNode;//initialize with the current state
TimerCallback* tcbPtr;///pointer to insert timers for each node

/******
Now update the state of the 
channel based on the state 
machine
*******/
switch(channelStateAtNode){

		case e_floor:{
			OUTCON(m_myId,  " cPhy::TDE->floor", "***STATE M/C ERROR ***",3);
			break;
			}
		case e_dataPresent:{//this is great.. a correct data reception
			//do X2 (i am being anal, but there should never be a TD occuring
			if(TD.isSet){
//			cout<<"Tone detect event set.. canceling, but IT SHOULDNT HAVE BEEN SET!!"<<endl;
			TD.isSet=false;
			gTimerManager.RemoveTimer(TD.eventHandle);
			}
			//channel is totaly clear now
			channelStateAtNode= e_floor;

			//change AM to listen now -- the MAC has to explicity turn it of
			amPower->changeDevState(e_am_on, globalClock.globalTime);

			OUTCON(m_myId,  " cPhy::TDE->dataPresent", "",3);
			break;
			}
		case e_dataOverlap:{
			//this case has many different scenarios
			timeTicks CT= globalClock.globalTime;
			//first we check if there is any other data present after
			//this data ends
			if(DCCT > CT){//yes
				//in this case simply return to the orignal state
				channelStateAtNode= e_dataOverlap;
				
				//also change AM to receive now -- the MAC has to explicity turn it of
				amPower->changeDevState(e_am_rx, globalClock.globalTime);

				OUTCON(m_myId,  " cPhy::TDE->dataOverlap", ">>dataOverlap",3);						
			}
			else{//all data energy is absent
			//in this case change the return state to show that all data has finished on the
			//data channel, but this was still a corrupted packet
			returnState = e_dataOverlapEnded;
						
			//change AM to listen now -- the MAC has to explicity turn it of
			amPower->changeDevState(e_am_on, globalClock.globalTime);

							
			//now check if any tone energy remains
				if(TCCT> CT){//yes there is some tone energy left
				//do Y
				
				//first check if the energy isnt due to our own tone transmission
					if(toneTxing==false){//In this case we have to setup a new TD event
						eventInformation.timeLeft_=(timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK);
						eventInformation.eType_=e_mac_toneDetected;
						eventInformation.nodeId_=m_myId;
						tcbPtr = new Event(eventInformation);
						TD.eventHandle = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle 
						TD.isSet=true;
					}
				//transit to the tone OVerlap state (eventhough there might not be any overlap in reality)
				//as it doesnt matter in the transition
				channelStateAtNode= e_toneOverlap;


				OUTCON(m_myId,  " cPhy::TDE->dataOverlap", ">>toneOverlap",3);						
				}
				else{//no type of energy remians on the channel
					//do X2
					if(TD.isSet){			
						TD.isSet=false;
						gTimerManager.RemoveTimer(TD.eventHandle);
					}
					channelStateAtNode=e_floor;
					OUTCON(m_myId,  " cPhy::TDE->dataOverlap", ">>floor",3);						
				}


			}

			break;
			}
		case e_tonePresent:{
			OUTCON(m_myId,  " cPhy::TDE->tonePresent", "***STATE M/C ERROR ***",3);
			break;
			}
		case e_toneOverlap:{
			OUTCON(m_myId,  " cPhy::TDE->toneOverlap", "***STATE M/C ERROR ***",3);			
			break;
			}
		case e_dataToneOverlap:{
			//check the state of tone energy after this event
			if(TCCT > globalClock.globalTime){//there is still tone energy present
				//do Y
				//first check if the energy isnt due to our own tone transmission
				if(toneTxing==false){//In this case we have to setup a new TD event
					eventInformation.timeLeft_=(timeTicks)(TONE_DETECT_TIME*SEC_TO_TIMETICK);
					eventInformation.eType_=e_mac_toneDetected;
					eventInformation.nodeId_=m_myId;
					tcbPtr = new Event(eventInformation);
					TD.eventHandle = gTimerManager.AddTimer(eventInformation.timeLeft_, tcbPtr);//keep the handle 
					TD.isSet=true;
				}
				//transit to the tone OVerlap state (eventhough there might not be any overlap in reality)
				//as it doesnt matter in the transition
				channelStateAtNode= e_toneOverlap;
				OUTCON(m_myId,  " cPhy::TDE->dataToneOverlap", ">>toneOverlap",3);						
			}
			else{//no type of energy remians on the channel
				//do X2
				if(TD.isSet){			
					TD.isSet=false;
					gTimerManager.RemoveTimer(TD.eventHandle);
				}
				channelStateAtNode=e_floor;
				OUTCON(m_myId,  " cPhy::TDE->dataToneOverlap", ">>floor",3);						
			}
				

			//also change AM to listen now -- the MAC has to explicity turn it of
			amPower->changeDevState(e_am_on, globalClock.globalTime);

			break;
			}
		default:
			OUTCON(m_myId,  " cPhy::TDE->UNKNOWNSTATE", "",3);
			break;

	}

/**********/
if(wuPower->currentState !=e_wakeup_off)//doing this b/c this is the way the power consumtion of the wakeup rx will change
	wakeUpOn(globalClock.globalTime);
					
	
return returnState;
}

void cPhy::wakeUpOn(timeTicks currTime){//MAC interface to basically turn on the WU and put it in either idle or rx state
//the assumption is that if their is data or tone energy on the channel, then if the wu is not OFF, it will take up as much
//energy as if it was reciving a tone
	if(TCCT > currTime || DCCT > currTime){
		if(wuPower->currentState !=e_wakeup_on) 
			wuPower->changeDevState(e_wakeup_rx,currTime);	
		}	
	else{
		if(wuPower->currentState !=e_wakeup_on)
			wuPower->changeDevState(e_wakeup_on,currTime);//no energy  (tone or data) present to keep the wu in receinvg state
		}
}
