/*
 * Author: Affan Syed <asyed@usc.edu>
 *   phy.h: The physical layer abstraction 
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */

#ifndef __PHY_H_
#define __PHY_H_

#include "defines.h"
#include "channel.h"
#include "power.h"


struct toneDetectEvent{
	bool isSet;//tells if the tone detect event is set or not
	handle eventHandle;//handle to the event in our simulation 
};
class cPhy{


	handle hdlToneEnd,hdlDataEnd;//these handles keep track of tone and data rx end events
	handle hdlToneTxEnd, hdlDataTxEnd;//while these two handle the end event when data is being transmitted 


	
	
	timeTicks toneStartTime;//the effective start time of the tone on the receiver circuitry
	timeTicks toneStartTxTime,dataStartTxTime;//these are the transmit times
	timeTicks DCCT;///data Channel Clear Time	(global)
	timeTicks TCCT;///Tone Channel Clear Time	(global)
	bool dataRxing,toneRxing;//indicators if phy is currently receiving some tone or data
	bool dataTxing,toneTxing;//indicators if phy is currently transmitting some tone or data
	bool tonePresent;// this indicates there is tone but since we are tx we  have our receiver turned off and wont be able to use it to energize our receive ciruitry
	short m_myId;
	channelState channelStateAtNode;///keep info about the state of the data channel

	cAcousticModemPower* amPower;
	cWakeupPower* wuPower;
	 float m_toneCancelProb;//used in dropping/ignoring tones
  public:
	channelState getDataChannelState(){return channelStateAtNode;};  
	toneDetectEvent  TD;
  	uint_32 ownToneCollision, ownDataCollsion; //this is the # of tone collisions that happened when we were transmitting
	uint_32 overlapToneCollision, overlapData;//this is # of tone overlaps that cant be diambiguated
  	cPhy(short Id,cAcousticModemPower* amPwr,cWakeupPower* wuPwr, float toneCancelProb):m_myId(Id),m_toneCancelProb(toneCancelProb){
  		
  		hdlDataEnd=hdlToneEnd=NULL;
  		toneStartTime=TCCT=DCCT=0;
  		toneStartTxTime=dataStartTxTime=0;
  		dataRxing=dataTxing=toneRxing=toneTxing=false;
  		tonePresent=false;
  		ownToneCollision=overlapToneCollision=0;	
		ownDataCollsion=overlapData=0;		
		
		channelStateAtNode=e_floor;
		TD.isSet=false;		

		amPower=amPwr;
		wuPower=wuPwr;
  	};


	bool resetPhy(){
		hdlDataEnd=hdlToneEnd=NULL;
  		toneStartTime=TCCT=DCCT=0;
  		toneStartTxTime=dataStartTxTime=0;
  		dataRxing=dataTxing=toneRxing=toneTxing=false;
  		tonePresent=false;
  		ownToneCollision=overlapToneCollision=0;	
		ownDataCollsion=overlapData=0;
		
		channelStateAtNode=e_floor;
		TD.isSet=false;
		return true;
		};
	void printsStats(FILE* outFile){
		fprintf(outFile,"Phy layer statistics (not detectable to MAC):\n");
		fprintf(outFile,"Own collisions (while transmitting): Tone = %d \t Data = %d \n", ownToneCollision, ownDataCollsion);
		fprintf(outFile,"Overlaps (while receiving): Tone= %d \t Data=%d \n",overlapToneCollision, overlapData);

		};
	channelState PhySendTone();//TTS	
	channelState PhySendData(Packet& pktTx);//DTS

	void wakeUpOn(timeTicks currTime);//MAC interface to basically turn on the WU and put it in either idle or rx state
	void wakeUpOff(timeTicks currTime){//MAC interface to basically turn off WU
		if(wuPower->currentState !=e_wakeup_off)//DONT change state is the state was previously the same
			wuPower->changeDevState(e_wakeup_off, currTime);
	}
	void amOff(timeTicks currTime){//MAC interface to turn off the acoustic modem
		if(amPower->currentState !=e_wakeup_off)
			amPower->changeDevState(e_am_off, currTime);

	};
	void amOn(timeTicks currTime){//MAC interface that is never called b/c of our trick of combining wu tone and data

		if(DCCT <= currTime){	
			if(amPower->currentState !=e_wakeup_on)
				amPower->changeDevState(e_am_on, currTime);
		}
		else{//if there is data on the channel turn it to rx mode
			if(amPower->currentState !=e_wakeup_rx)
				amPower->changeDevState(e_am_rx, currTime);
		}

	};
	bool amRx(timeTicks currTime){//MAC interface that is never called b/c of our trick of combining wu tone and data
		if(amPower->currentState ==e_am_on){
			amPower->changeDevState(e_am_rx, currTime);
			return true;
			}
		return false;

	};
	channelState PhyToneTxEnds(EventInfo eventInformation);//TTE
	channelState PhyDataTxEnds(EventInfo eventInformation);//TDE
	
//These next two should directly be invoked to handle events from the event queue.
	channelState PhyToneRxStart(EventInfo eventInformation);//RTS
	channelState PhyDataRxStart(EventInfo eventInformation);//RDS

	channelState PhyToneRxEnds(EventInfo eventInformation);//RTE
						
	channelState PhyDataRxEnds(EventInfo eventInformation);//RDE

};

#endif
