/*
 * Author: Affan Syed <asyed@usc.edu>
 *   power.cpp: Here all the device power characteristics are defined
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */
 #include "power.h"
 
extern cVTime globalClock;

 bool cWakeupPower::consumeEnergy(int state, timeTicks timeInState){

	switch(state){

		case e_wakeup_off:
			timeOff+=timeInState;
			energyConsumed += offPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
		case e_wakeup_on:
			timeOn+=timeInState;
			energyConsumed += onPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
		case e_wakeup_rx:
			timeRx+=timeInState;
			energyConsumed+=rxPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
		case e_wakeup_tx:
			timeTx+=timeInState;
			energyConsumed+=txPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
		default:
			if(DEBUG_LEVEL_CON >= 3)
			OUTCON(-1,"cWakeupPower::consumeEnergy","Unknown state",2);
			return false;
			break;
			
		}
	

 }

 bool cAcousticModemPower::consumeEnergy(int state, timeTicks timeInState){

	switch(state){

		case e_am_off:
			timeOff+=timeInState;
			energyConsumed += offPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
			
		case e_am_on:
			timeOn+=timeInState;
			energyConsumed += onPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;

		case e_am_rx:
			timeRx+=timeInState;
			energyConsumed+=rxPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
		case e_am_tx:
			timeTx+=timeInState;
			energyConsumed+=txPower*((double)timeInState/SEC_TO_TIMETICK);
			return true;
			break;
		default:
			if(DEBUG_LEVEL_CON >= 3)
			OUTCON(-1,"cAcousticModemPower::consumeEnergy","Unknown state",2);
			return false;
			break;
			
		}
	

 }
bool cAcousticModemPower::finalize(){//wrap up the power spent in finalize state
	this->changeDevState(e_am_off, globalClock.globalTime);//turn off the modem 
	return true;

}


bool cWakeupPower::finalize(){//wrap up the power spent in finalize stat
	this->changeDevState(e_wakeup_off, globalClock.globalTime);//turn off the wu circuit
	return true;
}
