/*
 * Author: Affan Syed <asyed@usc.edu>
 *   power.h: Here all the device power characteristics are declared along with the base class 
 *   that does power bookkeeping.
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */

#ifndef _DEVPOWER_H_
#define _DEVPOWER_H_


#include "defines.h"



class cDevPower {



public:
double energyConsumed;//in mJ
double onPower;//in mW THIS REPRSENTS WHEN THE DEVICE IS TURNED ON AND NOT BEING ACTIVE i.e. idle state of radio
timeTicks timeOn;//keeps track of time that you remained on. 

double offPower;//ion mW
timeTicks timeOff;//keeps track of time that you remained off. 

int currentState; //current state the device is in
timeTicks stateStartTime;//time at which the current state was started.

/***
Other than the above two generic power consuming states,
device specific time and power consumption values should
be defined in the power device below
***/
	cDevPower(){
		energyConsumed=0;
		timeOn=timeOff=0;		
	};
	virtual ~cDevPower(){};
//the first argument will be reinterpreted by each device as an enum describing its different operation states
	virtual bool consumeEnergy(int state, timeTicks timeInState)=0;
	virtual bool finalize()=0;
	 bool changeDevState(int newState, timeTicks currentTime){
		if(currentTime >= stateStartTime){
		consumeEnergy(currentState, (currentTime-stateStartTime));
		currentState=newState;
		stateStartTime=currentTime;
		return true;
		}
		else
			return false;
	}
	double getEnergyConsumed(){return energyConsumed;};
	virtual void printTimeInStates()=0;

	
};



 
class cWakeupPower: public cDevPower{
public:
double txPower;//when txmiting 
timeTicks timeTx;

double rxPower;//is this different from idle power?
timeTicks timeRx;


cWakeupPower(){
	onPower=WAKEUP_OFF_POWER;
	offPower=WAKEUP_OFF_POWER;
	txPower= WAKEUP_TX_POWER;
	rxPower=WAKEUP_RX_POWER;//so right now out off and receive power for the wakeup modem are the same. 
	timeRx=timeTx=0;
	currentState=e_wakeup_on;//start here so we dont have to worry about turning it on
	stateStartTime = (timeTicks) 0;//from the start of the simulation!
	};
bool consumeEnergy(int state, timeTicks timeInState);
	void printTimeInStates(){
			printf("%f\t\t%f\t\t%f\t\t%f\t\t",(timeOff/SEC_TO_TIMETICK),(timeTx/SEC_TO_TIMETICK),
				(timeRx/SEC_TO_TIMETICK),(timeOn/SEC_TO_TIMETICK));
			};

bool finalize();

};


class cAcousticModemPower: public cDevPower{

public:
double txPower;//when txmiting 
timeTicks timeTx;

double rxPower;//is this different from idle power?
timeTicks timeRx;


cAcousticModemPower(){
	onPower=AM_OFF_POWER;//for now idle power is same is off power
	offPower=AM_OFF_POWER;
	txPower=AM_TX_POWER;
	rxPower=AM_RX_POWER;
	timeRx=timeTx=0;
	currentState=e_am_off;//data circuit remians off
	stateStartTime = (timeTicks) 0;//from the start of the simulation!
	};

bool consumeEnergy(int state, timeTicks timeInState);
void printTimeInStates(){
			printf("%f\t\t%f\t\t%f\t\t%f\t\t",(timeOff/SEC_TO_TIMETICK),(timeTx/SEC_TO_TIMETICK),(timeRx/SEC_TO_TIMETICK),(timeOn/SEC_TO_TIMETICK));
		};

bool finalize();

};


#endif
