/*
 * Author: Affan Syed <asyed@usc.edu>
 *   simNode.h: Declaration of node specific structures and their properties
*		Also defines the "node" used in our simulation. 
 *
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */
#ifndef _SIMNODE_H_
#define _SIMNODE_H_
#include "TDMA.h"
#include "CSMA.h"
#include "defines.h"
#include "lineReg.h"
#include "STLohi.h"
#include "UTLohi.h"
#include "Aloha.h"
#include "slottedAloha.h"


/* Supporting Data structures used in the simulation */


class cNode{
	public:
		//Affan: here I am assuming that the directory to which you write will be provided , already opened as a pointer to each node instance - since this is a discrete 
		//event driven structure, there will be no race conditions
		cNode(Coordinate location, macParameter macParams, short nodeId, FILE* outputFile,
		trafficInfo traffic):outFile(outputFile), myLocation(location), myId(nodeId),myMacParams(macParams), myTraffic(traffic){
				macInstantiated =false;
				m_appPktsReceived=0;
				m_burstCount=0;

		};
		~cNode(){ 
			if(macInstantiated==true)
				delete MAC;
			};

		cNode(const cNode& orig);
		cNode& operator= (const cNode& orig);
		bool hdlDeaf;

		//the return value is to suggest to the timer whether it should set another value or not
		bool NodeTransmitEnd(EventInfo& rcvdPkt);///MAC layer transmission of  packet completed
		bool NodeAppTransmit(EventInfo& txPkt);///Traffic generted - Try and send it using the MAC
		void resetNode();///use this function to reset a node state... used only by the simulator to rerun multiple times
		void initNode();//use this to create the correct file out put and initialize all node variables. 
		void informNodeCount(int nodeCount){MAC->setNodeCount(nodeCount);}
		
		void printStats(){
		//	cout<<myId<<" ("<<myLocation.x<<","<<myLocation.y<<")  ";
			MAC->printStats();
			};
		void finalizeEnergy(){
			MAC->amPower.finalize();
			MAC->wuPower.finalize();
			};
	
		void closeNode(){
		//Affan: Add any code to do end of simulation wrap up
		};
		bool operator==(const short id) const { return (this->myId ==id);}; //this is used in the STL find function

		FILE* outFile;
		cMAC* MAC; //this is the pointer to the MAC class containted in side the node. Should be instantiated with a 
		short myId;// the unique node id kept for purposes of identification
		uint_32 m_appPktsReceived;//keeps track of App pkts received by the app layer
		trafficInfo myTraffic;
		macParameter myMacParams;
		
		//This func is used to compute the euclidean distance b/w two nodes
		double getDistance(const cNode &node );

		private:
		bool macInstantiated;
		
		uint_8 m_burstCount;
		
		cNodeClock myClock;
		Coordinate myLocation;
		nodeState myState;
		
		cLinearRegression linRegParam; ///the class that holds the linear regression params
		queue<EventInfo> pktQueue;//This keeps a queue of packets that will be maintained at a local node
	
		

		///this function returns the time corrected from the nodes clock value as per the TimeSync params calculated
		//by linear regression. Option to use either only the calculated slope (withOffset=false) or both slope and offset.
		timeTicks GetCorrectedTime(bool withOffset);

		//Run the Estimation of clock skew at the end of phase1
		bool  EstimateBeaconClock();

		
};


#endif
