
/*
 * Author: Affan Syed <asyed@usc.edu>
 * SimTime.h:
 *   
 **This class is to abstract the notion of time in the simulator package
 *Specific notion of time can be implemented by inheriting from this class
 * The inherited class should have the notion in the form of attributes that will 
 * have to be independently interpreted by the user.
 *
 * If you need to pass parameters to your Time,
 * pass them in the constructor of your subclass.
 *
 * any data allocation in subclass should be dealt by the subclass
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */
 #ifndef _SIMTIME_H_
#define _SIMTIME_H_
 
 
#define MAXVALUE    0x7ffffff // no timer present 


 typedef long handle;
 
 typedef  unsigned long timeTicks; ///the basic unit of time increment
 

 class cSimTime{
 public:
	 cSimTime(){};
	 virtual ~cSimTime(){};
 
	 //the current virtual time 
	 virtual void getTime() =0;
	 
	 //return the current time in the cSimTime object
	 virtual cSimTime* returnCurrentTime()=0;
	 
	 //return -1,0,1  for < == >
	 virtual short compareTime( cSimTime* compareWith)=0;
 
	 //add to this time object a specific # of virtual ticks
	 virtual void addTime(timeTicks ticks)=0;
 
	 //set time arbitrarily to the value passed
	 virtual void setTime(timeTicks ticks)=0;
 
	 //subtract from this time object y  and return their value... return value interpreted by implementor of subclass
	 virtual cSimTime* subTime(cSimTime* y)=0 ;

	//assign the time value of one to the other using a pointer variable
	virtual void assign(cSimTime *y)=0;
	 
	 //generic printing funtion
	 virtual void print()=0;
	 
 };
#endif  
