/*
 * Author: Affan Syed <asyed@usc.edu>
 * simulation.cpp : The  simulator's skeleton
 *   
 *Description: This application creates the network, and provides the stimulus to the 
*			Network by setting a single beacon broadcast at the beacon node.This is achieved
*			using the cSimulator Class. Everything else	follows according to the 
*			state maching transition. Simulation ends when no more events are in the event queue
 *									
 * Copyright (c) 2005 the University of Southern California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF SOUTHERN CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
 * UNIVERSITY OF SOUTHERN CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * THE UNIVERSITY OF SOUTHERN CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF SOUTHERN CALIFORNIA HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 *
 */
#include "simulation.h"
cVTime globalClock;///Our global clock!
vector<cNode> gNodeVector;///vector that will keep track of all the nodes in the simulation
vector<nodeInfo> gNodeInfoVector;
Timers gTimerManager;
extern 	cChannel gChannel;

#define PI 3.1246

//initialize the simulator using a configuration file
bool cSimulator::Initialize(string cfgFileName, double range, double frac){

	ifstream cfgFile;
	string outFileName= directory_+"trace.out";
	int simSeed;
	int numNodes;

	cfgFile.open(cfgFileName.c_str());
	
	if (!cfgFile.good() )
		{
		cerr<<"CRITICAL ERROR: Configuration file does could not be found"<<cfgFileName<<endl;
			return false;
		}

	Coordinate location;
//	float skew;
//	timeTicks offset;
	string macType; //mac_type currenty st or ut lohi	
	string strTrafficType; //traffic currenty 
	short nodeId;
	macParameter mParam;
	trafficInfo mTraffic;
	mParam.maxRange=range;
	mParam.fracDelay=frac;
	gChannel.setMaxRange(range);

	fptr= fopen(outFileName.c_str(),"w+");
	
	
	//now lets start parsing
	string token;
	string tempString;	

/****
	FORMAT of sim.cfg file
L1:	integer SEED
L2:	mac type as either STLohi or UTLohi (Case sensitive).		
L3:   traffic type  =	manual, CBR, poisson, saturated (Case sensitive).
	If CBR or poisson, L4 will have traffic parameter (float)
L4:   SIMTIME in sec(s) integer
L5:	PKTLEN in bytes (integer)
L6:   NUM_NODES the number of nodes to instantiate
L7:   topology type random OR manual
L7a---: Node Id and locations get node id's and their X,Y coordinates (if manual)
L7a+numNodes (if manual): 	if(macType is manual - for test purposes), get application transmit events for nodes

	***/

	cfgFile>>simSeed;///seed value
	getline(cfgFile,tempString);//ignore everthing else and goto next line

	srand(simSeed);
	
	cfgFile>>macType;///mac_type
		
		if ("STLohi" == macType){
				mParam.myMacType= e_STLohi;
		}
		else if("UTLohi" == macType ){
				mParam.myMacType = e_UTLohi;
		}
		else if("ALOHA" == macType ){
				mParam.myMacType = e_ALOHA;
		}
		else if("SALOHA" == macType ){
				mParam.myMacType = e_SALOHA;
		}
		else if("TDMA" == macType ){
				mParam.myMacType = e_TDMA;
		}
		else if("CSMA" == macType ){
				mParam.myMacType = e_CSMA;
		}
		else{			
			cerr<<"ERROR:unrecognized MAC type"<<endl;
			return false;
			
		}
getline(cfgFile,tempString);//ignore everthing else and goto next line

cfgFile>>strTrafficType; //get the rate at which traffic is being generated. 
if("manual"==strTrafficType){
	mTraffic.eType=e_manual;
}
else if("CBR"==strTrafficType){
	mTraffic.eType=e_CBR;
	getline(cfgFile,tempString);//ignore everthing else and goto next line
	cfgFile>>mTraffic.parameter;
}
else if("poisson"==strTrafficType){
	mTraffic.eType= e_poisson;
	getline(cfgFile,tempString);//ignore everthing else and goto next line
	cfgFile>>mTraffic.parameter;
}
else if("saturated"==strTrafficType){
	mTraffic.eType=e_saturated;
}
else if("bursty"==strTrafficType){
	mTraffic.eType=e_bursty;
	getline(cfgFile,tempString);//ignore everthing else and goto next line
	cfgFile>>mTraffic.parameter;//this is theavg poisson load, will be bursty at 1/10 the scale
}
else{

cerr<<"ERROR:unrecognized traffic  type"<<endl;
			return false;
}



getline(cfgFile,tempString);//ignore everthing else and goto next line

cfgFile>>simTimeInSec; //get the time for which simulation should be run in terms of seconds.
MAX_SIMTIME= (timeTicks)(simTimeInSec*SEC_TO_TIMETICK);//convert that in to max number of time 
//ticks for which the sumualtion should run 

getline(cfgFile,tempString);//ignore everthing else and goto next line
/*
cfgFile>>tempString;///mac_type
		
		if ("true" == tempString){
				mParam.hdlBiDeafness= true;
		}
		else if("false" == tempString ){
				mParam.hdlBiDeafness = false;
		}
		else{			
			cerr<<"ERROR:unrecognized handleDeafness type"<<endl;
			
		}
		
getline(cfgFile,tempString);//ignore everthing else and goto next line
*/
		
mParam.hdlBiDeafness = false;



mTraffic.ePktType = e_constant;//NOTE: can make the packet length variable by changing this as well. 


//L5: toneCancelProb
cfgFile>>mParam.toneCancelProb;
getline(cfgFile,tempString);


//L6: PKTLENT

cfgFile>>mTraffic.pktLength;
getline(cfgFile,tempString);

//L7: NUM_nodes 
cfgFile>>numNodes;
getline(cfgFile,tempString);//ignore everthing else and goto next line

//L8: now parse lines for different nodes and their locations
//first figure out if they are manual or random
cfgFile>>tempString; //TOKEN telling if the node positioning is manual or random

if("manual"==tempString){

/*
	Cfg file format should have the following structure
	<node_id> <locX,locY> 
e.g.   0	4,5	

	*/
getline(cfgFile,tempString);//go to the start of next line	

	while(!cfgFile.eof() && numNodes!=0){

		cfgFile>>nodeId;
		
		
		cfgFile>>location.x;
		cfgFile.ignore(); //ignore the comma sepration
		cfgFile>>location.y; 

		getline(cfgFile,tempString);//go to the start of next line\
		
		AddNode(location,mParam,nodeId,fptr,mTraffic);//add this node
		numNodes--;		
	}
}
else if("random"==tempString){
	getline(cfgFile,tempString);//go to the start of next line
	for(int id=0; id<numNodes;id++){
		location.x=RAND_UNIFORME*MAX_X;
		location.y=RAND_UNIFORME*MAX_Y;
		AddNode(location,mParam,id,fptr,mTraffic);//add this node
		}

}
else if("circular"==tempString){

	getline(cfgFile,tempString);//go to the start of next line
	location.x=0;
	location.y=0;
	int id=0;
	AddNode(location,mParam,id,fptr,mTraffic);//add the node with id 0 over here first
	//now deploy in a radial formation around it
	double radius = mParam.maxRange;
	double maxAngle=2*PI;
	double angle=0;

	for(id=1; id< numNodes;id++){ //unifromly distribute nodes in r and theta
		angle=RAND_UNIFORME*maxAngle;		
		location.x=cos(angle)*RAND_UNIFORME*radius;
		location.y=sin(angle)*RAND_UNIFORME*radius;
		AddNode(location,mParam,id,fptr,mTraffic);//add the node with id 0 over here first
	}

}

		
//L(8+numNodes): 
	if(mTraffic.eType==e_manual){
		int numOfTxEvents;
		short destId;
		float eventTime;
		TimerCallback* tcbPtr;///pointer to insert timers for each node
		EventInfo timerInfo;

		
		cfgFile>>numOfTxEvents;
		getline(cfgFile,tempString);//go to the start of next line\

		
	
		timerInfo.eType_ = e_tx;//set up an application layer transmission event	
		timerInfo.packet_.msgType_ = e_msg_data;//obviously the application has only data pkts to send
		timerInfo.packet_.length_ =mTraffic.pktLength;

			while(!cfgFile.eof()&& numOfTxEvents!=0){
				/*
				Transmit event have the following structure
				<src_id> <dest_id> <eventTime> in SECONDS (upto 3 decimal pts)
				e.g.   0	1 0.01
				*/
				cfgFile>>nodeId;
				cfgFile>>destId;
				cfgFile>>eventTime;

				timerInfo.nodeId_=nodeId;
				timerInfo.timeLeft_ = (timeTicks)(eventTime*SEC_TO_TIMETICK);// set up the first packet transmission for this 	
				timerInfo.packet_.destId_ = destId;
				timerInfo.packet_.sourceId_ =timerInfo.nodeId_;		
				tcbPtr = new Event(timerInfo);
				//insert the timer into the timer manager
				gTimerManager.AddTimer(timerInfo.timeLeft_, tcbPtr);					
				numOfTxEvents--;
				getline(cfgFile,tempString);//go to the start of next line
		}

	}



	cfgFile.close();//make sure you close the file!
	return true;
}

void cSimulator::AddNode(Coordinate location,macParameter macParams,short nodeId, FILE*  fptr, trafficInfo traffic){
		
	cNode* temp;
	nodeInfo* info;
	nodeCount++;
    
	temp = new cNode(location,macParams,nodeId,fptr, traffic);
	info = new nodeInfo(nodeId,location);
	
	gNodeVector.push_back(*temp);
	gNodeInfoVector.push_back(*info);//also add the info node to be used by the channel

	gNodeVector.back().initNode();//initialize this new node formed!!

	delete temp;
	delete info;

}


void cSimulator::SimStart(){
	
	/***
	For the underwater mac simulator, generate first transmit time 
	for each node based on a poisson arrival time and 
	then at each transmission generate the next transmit event to simulate traffic source. 
	***/
	int i;
	TimerCallback* tcbPtr;///pointer to insert timers for each node
	EventInfo timerInfo;
	int nodeIter;//we use this go through all nodes in simulation and seed transmission at them
		
	//IMP:: Set the global time to zero after this so that we can use restart our simulation any number of times
	globalClock.globalTime = 0;			
	
	for(nodeIter =0; nodeIter < nodeCount; nodeIter++){
		//inform the # of nodes in the simulation to each node

		gNodeVector[nodeIter].informNodeCount(nodeCount);
		/****
		Set the values for an event 				
		***/			

		timerInfo.eType_ = e_tx;//set up an application layer transmission event
		timerInfo.nodeId_=gNodeVector[nodeIter].myId;//set the event for this node
//Traffic Model: Add if(check node's traffic model).
		if(gNodeVector[nodeIter].myTraffic.eType ==e_poisson)
			timerInfo.timeLeft_ = (timeTicks)(random_exp(gNodeVector[nodeIter].myTraffic.parameter)*SEC_TO_TIMETICK);// set up the first packet transmission for this 
		else if(gNodeVector[nodeIter].myTraffic.eType ==e_CBR)
			timerInfo.timeLeft_ = (timeTicks)((gNodeVector[nodeIter].myTraffic.parameter)*SEC_TO_TIMETICK);// set up the first packet transmission for this 
		else if(gNodeVector[nodeIter].myTraffic.eType ==e_saturated)
			timerInfo.timeLeft_ = (timeTicks)(0);// set up the first packet transmission for this 			
		else if(gNodeVector[nodeIter].myTraffic.eType ==e_bursty)
			timerInfo.timeLeft_ = (timeTicks)(random_exp(gNodeVector[nodeIter].myTraffic.parameter/BURST_PARAM)
			*SEC_TO_TIMETICK);// set up the first packet transmission for this at a rate 1/10 of the avg
	
		//otherwise its a manual entry so we dont want to worry about it.			
		
		/****
		Now that we have set all the info needed for the event... create the data packet
		***/			
		int dest;
		//HACK!!! with ALOHA test I just want to send it always to center node that must be the first addition to the node list
/*HACK REMOVED
		if(gNodeVector[nodeIter].myMacParams.myMacType==e_ALOHA ||
			gNodeVector[nodeIter].myMacParams.myMacType==e_SALOHA){
			dest=0;
		}
		else{//otherwise randomly choose another destination
			do{//find a destination index other than me	
				dest = (int) (gNodeVector.size()* RAND_UNIFORME);
				}while(dest ==nodeIter && dest <=nodeCount );
		}	

		*/
		do{//find a destination index other than me	
				dest = (int) (gNodeVector.size()* RAND_UNIFORME);
				}while(dest ==nodeIter && dest <=nodeCount );
	
		timerInfo.packet_.destId_ = gNodeVector[dest].myId;
		timerInfo.packet_.msgType_ = e_msg_data;//obviously the application has only data pkts to send
		timerInfo.packet_.sourceId_ =gNodeVector[nodeIter].myId;

		if(gNodeVector[nodeIter].myTraffic.ePktType ==e_constant)
			timerInfo.packet_.length_ =gNodeVector[nodeIter].myTraffic.pktLength;
		else if(gNodeVector[nodeIter].myTraffic.ePktType==e_uniform)
			timerInfo.packet_.length_ =(int)(gNodeVector[nodeIter].myTraffic.pktLength*RAND_UNIFORME);
		
	
		tcbPtr = new Event(timerInfo);//caution: make sure you delete the tcbPtr
		//insert the timer into the timer manager
		gTimerManager.AddTimer(timerInfo.timeLeft_, tcbPtr);

	}
		
	EventLoop();		
	//now that the simulation is about to end.. reset all nodes 
	for (int j=0; j<(int) gNodeVector.size(); j++){ //loop through all nodes and reset them 
		gNodeVector[j].printStats();
		gNodeVector[j].resetNode();							 
	cout<<endl<<endl;
	}


		for (i=0; i< (int)gNodeVector.size(); i++){ //loop through all nodes and wrap up things 
			  gNodeVector[i].closeNode();
			}
}

/****
This executes the event in a sequential loop,
exhausting all possible events before exiting.
****/
void cSimulator::EventLoop(){

	cVTime  nextEventTime;
	
//	fprintf(stderr,"Starting Event loop at global time =%d\n", globalClock.globalTime);
	
	
	while (1) {//there will always be an event in the queue due to the way traffic is generated
		gTimerManager.NextTimerTime(&nextEventTime);//this will return the time left till the next event 
		//fprintf(stderr,"Next Event Time Left = %d\n", nextEventTime.m_localTime);

		if (0 == nextEventTime.m_localTime ) { //this is a cVTime specific interpretation of return val by subTime()
		        // Time to execute the event/timer at the head of the queue 
		        gTimerManager.ExecuteNextTimer();
			continue;
		}
		if (nextEventTime.m_localTime == MAX_TIMER_VALUE){
		        // There are no timers in the event queue 
		        
//			fprintf(stderr,"Simulation ends at global time %d\n", globalClock.m_localTime);
		        break;
		}
		  
		//if not we should advance the global time to the value of the next event and execute it
		globalClock.advance_gTime(nextEventTime.m_localTime);
		if(globalClock.globalTime > MAX_SIMTIME){//if the next event is going to happen after the simulation time then we end simulation now
			globalClock.globalTime =MAX_SIMTIME;
			break;
			}
		// We have reached the point where we have to execute the event at head of queue
	        gTimerManager.ExecuteNextTimer();
		// Execute all timers that have expired.
		gTimerManager.NextTimerTime(&nextEventTime);
		while(0 == nextEventTime.m_localTime){
		       //Timer at the head of the queue has expired 
		       gTimerManager.ExecuteNextTimer();
			gTimerManager.NextTimerTime(&nextEventTime);
			}
		
	}
	
	
	return;
}


